#!/usr/bin/env bash
{
set -o nounset

SCRIPT_DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"
readonly SCRIPT_DIR
SCRIPT_FILENAME="$(basename "${BASH_SOURCE[0]}")"
readonly SCRIPT_FILENAME

. "${SCRIPT_DIR}/include.sh"

readonly SCRIPT_NAME="${PRODUCT_NAME}-backup"
readonly SCRIPT_VERSION="3.11.10"

readonly LAUNCH_DATETIME_FOR_FILE="$(date "+${DATETIME_FORMAT_FOR_FILE}")"

SUDO_CMD=''

SCRIPT_USAGE="$(cat <<EOF
Usage:
  ${SCRIPT_FILENAME} [options]

Options:
  -v
      Show script version.
  --legal
      Show legal information.
  -h, --help
      Show help.
EOF
)"
readonly SCRIPT_USAGE

backup_file_or_dir() {
  local filedir="$1"
  local backup_dir="$2"

  local parent_dir
  parent_dir="$(dirname "${filedir}")"

  if ! mkdir -p "${backup_dir}${parent_dir}"; then
    return 1
  fi

  if ! ${SUDO_CMD} cp -Rp "${filedir}" "${backup_dir}${parent_dir}"; then
    return 2
  fi
}

check_and_backup() {
  local filedir="$1"
  local backup_dir="$2"
  local name="$3"

  if [ -f "${filedir}" ] || [ -d "${filedir}" ]; then
    echo "  Found ${name}, copy it: '${filedir}'..."
    backup_file_or_dir "${filedir}" "${backup_dir}"
    ok_or_die $? "  OK" "Failed to backup ${name}: '${filedir}'"
  else
    echo "  ${name} does not exist, do NOT copy: '${filedir}'"
  fi
}

make_backup_copy() {
  local backup_dir="$1"

  echo "Create backup directory: '${backup_dir}'..."
  if ! mkdir -p "${backup_dir}"; then
    echo "FAIL"
    return 1
  fi
  echo "OK"

  echo "Start backup..."

  local files
  files=()

  find_files() {
    local dir="$1"
    local name="$2"
    local exceptions=("${@:3}")

    local file_path
    for file_path in "${dir}"/*itgrix*; do
      if ! in_array "${file_path}" "${exceptions[@]}"; then
        files+=("${file_path}" "${name}")
      fi
    done
  }

  files+=("${APP_DIR}" "app directory")
  files+=("${CONFIG_DIR}" "config directory")
  files+=("${LOGROTATE_CONFIG_PATH}" "logrotate config file")
  files+=("${SERVICE_FILE_SYSTEMD}" "service file (systemd)")
  files+=("${SERVICE_FILE_INIT}" "service file (init)")
  files+=("${SUDOERS_FILE}" "sudoers file")
  files+=("${SUPERV_LOGROTATE_CONFIG_PATH}" "supervisor logrotate config file")
  files+=("${SUPERV_SERVICE_FILE_SYSTEMD}" "supervisor service file (systemd)")
  files+=("${SUPERV_SERVICE_FILE_INIT}" "supervisor service file (init)")

  find_files "${LOGROTATE_CONFIG_DIR}" \
      "additional(*) logrotate config file" \
      "${LOGROTATE_CONFIG_PATH}" "${SUPERV_LOGROTATE_CONFIG_PATH}"

  find_files "${SERVICE_DIR_SYSTEMD}" \
      "additional(*) service file (systemd)" \
      "${SERVICE_FILE_SYSTEMD}" "${SUPERV_SERVICE_FILE_SYSTEMD}"

  find_files "${SERVICE_DIR_INIT}" \
      "additional(*) service file (init)" \
      "${SERVICE_FILE_INIT}" "${SUPERV_SERVICE_FILE_INIT}"

  local i
  local file
  local name
  for (( i = 0; i < ${#files[@]}; i += 2 )); do
    file="${files[${i}]}"
    name="${files[$((i+1))]}"
    check_and_backup "${file}" "${backup_dir}" "${name}"
  done
}

make_backup() {
  local app_version="$1"

  echo
  echo "Check permissions for '${BACKUPS_DIR}' ..."
  check_dir_write_permissions "${BACKUPS_DIR}"
  ok_or_die $? "OK" "Permissions check failed: cannot write in '${BACKUPS_DIR}'"

  echo
  echo "Backup..."
  local app_version_escaped
  app_version_escaped="$(escape_version "${app_version}")"
  local backup_dir="${BACKUPS_DIR}/${PRODUCT_NAME}-${LAUNCH_DATETIME_FOR_FILE}-${app_version_escaped}"
  make_backup_copy "${backup_dir}"
  ok_or_die $? "OK, backup done" "Failed to make backup (code: $?)"
}

main() {

  for i in "$@"; do
    case "${i}" in
      -v)
        print_version
        exit
        shift
        ;;
      --legal)
        print_version
        echo
        print_legal
        exit
        shift
        ;;
      -h|--help)
        print_help
        exit
        shift
        ;;
      *) # unknown option
        die_usage "Unknown option: ${i}"
        ;;
    esac
  done

  echo -n "Start " && print_version

  local cur_eff_user
  local cur_eff_group
  cur_eff_user="$(id -un)"
  cur_eff_group="$(id -gn)"

  echo
  echo "Current effective user: '${cur_eff_user}:${cur_eff_group}'"

  echo
  echo "Check sudoers..."
  set_sudo_cmd "${cur_eff_user}"

  echo
  if [ -d "${BACKUPS_DIR}" ]; then
    echo "Directory for backups already exists, do not create: '${BACKUPS_DIR}'"
  else
    echo "Create directory for backups: '${BACKUPS_DIR}'..."
    mkdir -p "${BACKUPS_DIR}"
    ok_or_die $? "OK" "Failed to create directory for backups: '${BACKUPS_DIR}'"
  fi

  echo
  echo "Get app version..."
  local app_version
  app_version="$(get_bin_version "${BIN_PATH}")"
  echo "Got app version: '${app_version}'"

  make_backup "${app_version}"

  echo
  echo -n "Finish " && print_version
}
main "$@"
exit
}
