<?php

require_once __DIR__.'/Settings.php';
require_once __DIR__.'/Utils.php';
require_once __DIR__.'/crm.php';

define('ACTIONS_DIR', realpath(__DIR__.'/../actions'));

/**
 * Class Customizer
 *
 * Обеспечивает исполнение пользовательских кастомизаций.
 *
 * В основном содержит служебные методы,
 * пользователю необходимы только методы записи в лог.
 *
 * При запуске кастомизации пользователю будут доступны переменные:
 *   $action - название кастомизации (например, 'process_client_phone');
 *   $params - параметры, которые модуль Itgrix передал в кастомизатор;
 * а также псевдопеременная $this (экземпляр Customizer),
 * необходимая для логирования, например:
 *   $this->logInfo('...').
 *
 * В процессе выполнения может потребоваться чтение конфига и запись в лог.
 * Пути к конфигу и логу прописаны в settings.json.
 */
class Customizer {

    /**
     * Customizer constructor.
     *
     * @param string $action Название кастомизации
     * @param array  $params Ассоциативный массив параметров
     *
     * @throws Exception В случае ошибки
     */
    public function __construct($action, $params)
    {
        $this->action_ = $action;
        $this->params_ = $params;

        Utils::setOsTimeZoneAsDefault();
    }

    /**
     * Сделать запись уровня debug в лог
     *
     * В поле 'command' будет записано название кастомизации (action),
     * из которой вызвано логирование.
     * В поле 'call_id' будет ID звонка, если он содержится в параметрах кастомизации.
     *
     * Путь к логу прописан в settings.json.
     *
     * @param string $message Сообщение
     *
     * @throws Exception В случае ошибки
     */
    public function logDebug($message) {
        $this->log('debug', $message);
    }

    /**
     * Сделать запись уровня info в лог
     *
     * В поле 'command' будет записано название кастомизации (action),
     * из которой вызвано логирование.
     * В поле 'call_id' будет ID звонка, если он содержится в параметрах кастомизации.
     *
     * Путь к логу прописан в settings.json.
     *
     * @param string $message Сообщение
     *
     * @throws Exception В случае ошибки
     */
    public function logInfo($message) {
        $this->log('info', $message);
    }

    /**
     * Сделать запись уровня warning в лог
     *
     * В поле 'command' будет записано название кастомизации (action),
     * из которой вызвано логирование.
     * В поле 'call_id' будет ID звонка, если он содержится в параметрах кастомизации.
     *
     * Путь к логу прописан в settings.json.
     *
     * @param string $message Сообщение
     *
     * @throws Exception В случае ошибки
     */
    public function logWarn($message) {
        $this->log('warning', $message);
    }

    /**
     * Сделать запись уровня error в лог
     *
     * В поле 'command' будет записано название кастомизации (action),
     * из которой вызвано логирование.
     * В поле 'call_id' будет ID звонка, если он содержится в параметрах кастомизации.
     *
     * Путь к логу прописан в settings.json.
     *
     * @param string $message Сообщение
     *
     * @throws Exception В случае ошибки
     */
    public function logError($message) {
        $this->log('error', $message);
    }

    public function run() {
        $file = $this->makeActionFilename();

        if (! is_file($file)) {
            return array(
                'state' => 'disabled',
                'data' => sprintf("Action file does not exist or is not a regular file: '%s'", $file),
            );
        }

        return $this->runAction();
    }

//==============================================================================

    /** @var string Название кастомизации */
    protected $action_;

    /** @var array Ассоциативный массив с параметрами */
    protected $params_;

    protected function runAction() {
        $action = $this->action_;
        $params = $this->params_;

        return require $this->makeActionFilename();
    }

    protected function makeActionFilename() {
        return sprintf('%s/%s.php', ACTIONS_DIR, $this->action_);
    }

    /**
     * Вывести сообщение в лог
     *
     * В поле 'command' будет записано название кастомизации (action),
     * из которой вызвано логирование.
     * В поле 'call_id' будет ID звонка, если он содержится в параметрах кастомизации.
     *
     * @param string $level   Уровень лога (debug, info, warning, error)
     * @param string $message Сообщение
     *
     * @throws Exception В случае ошибки
     */
    protected function log($level, $message) {
        Utils::log($this->action_, $this->params_, $level, $message);
    }

}
