<?php

require_once __DIR__.'/Settings.php';

// Для PHP < 5.4.0
// По сути работать не будет, нужно только чтобы json_encode вообще выполнилась.
if (! defined('JSON_UNESCAPED_UNICODE')) {
    define('JSON_UNESCAPED_UNICODE', 256);
}

define('MAX_LOG_LEVEL_NAME_LENGTH', 7); // debug, info, warning (7), error

/**
 * Class Utils
 *
 * Реализует вспомогательные функции.
 * Они все статические.
 *
 * По необходимости функции читают данные из конфига модуля Itgrix.
 * При возникновении ошибки функции делают запись в лог модуля Itgrix.
 *
 * Пути к конфигу и логу прописаны в settings.json.
 */
class Utils {

    /**
     * Установить временную зону ОС как зону по умолчанию (до окончания работы кастомизатора).
     *
     * Функция попробует последовательно получить варианты временной зоны запуском команд:
     *   - timedatectl (поле Time zone)
     *   - date +%Z
     *   - date +%-:::z (только если зона целочисленная,
     *                  например, для +5 будет Etc/GMT-5, а зоны +10:30 и +8:45 не подходят)
     *
     * Если ни один из вариантов не будет распознан в PHP (DateTimeZone),
     * то по умолчанию будет установлена зона UTC.
     */
    public static function setOsTimeZoneAsDefault() {
        // Если display_errors = STDOUT (например, = On в настройках PHP),
        // и error_reporting разрешает уровень E_NOTICE (например, error_reporting = E_ALL & ~E_DEPRECATED),
        // то вывод ошибок date_default_timezone_set попадет в вывод кастомизации,
        // что ломает дальнейший парсинг модулем.
        //
        // Пример вывода кастомизации:
        // Notice: date_default_timezone_set(): Timezone ID '+05' is invalid
        // {"state":"success","data":[]}
        //
        // Чтобы этого избежать, отключим вывод сообщений уровня E_NOTICE на время выполнения этой функции.
        $errorReportingBefore = error_reporting();
        error_reporting($errorReportingBefore & ~E_NOTICE);

        $timedatectlZone = '';
        // Пример вывода:
        //   Time zone: Asia/Yekaterinburg (+05, +0500)
        exec("which timedatectl > /dev/null 2>&1 && timedatectl | grep 'Time zone' 2>/dev/null",
            $timezoneOutput, $exitCode);
        if (($exitCode === 0) &&
            (preg_match('/^[^\:]+\:[\ \t]*([^\ \t]+).*$/', $timezoneOutput[0], $matches) === 1)) {
            $timedatectlZone = $matches[1];
        }
        if (date_default_timezone_set($timedatectlZone)) {
            error_reporting($errorReportingBefore);
            return;
        }

        $dateZone = '';
        exec('date +%Z 2> /dev/null', $dateOutput, $exitCode);
        if ($exitCode === 0) {
            $dateZone = $dateOutput[0];
        }
        if (date_default_timezone_set($dateZone)) {
            error_reporting($errorReportingBefore);
            return;
        }

        $dateNumericZone = '';
        // Примеры вывода: +0, +5, -4, -10, +4:30, +10:30, +8:45
        exec('date +%-:::z 2> /dev/null', $dateNumericOutput, $exitCode);
        if ($exitCode === 0) {
            $dateNumericZone = $dateNumericOutput[0];
            if (preg_match('/^(\+|-)[0-9]{1,2}$/', $dateNumericZone) === 1) { // например, +0, +5, -10, но не +9:30
                // В PHP используются часовые пояса из базы IANA, они в POSIX-стиле, знак обратный.
                // Например, если зона +5, в PHP нужно установить Etc/GMT-5.
                // https://www.php.net/manual/ru/timezones.others.php
                // https://data.iana.org/time-zones/tzdb/etcetera
                if ($dateNumericZone[0] === '+') {
                    $dateNumericZone[0] = '-';
                } else {
                    $dateNumericZone[0] = '+';
                }
                $dateNumericZone = 'Etc/GMT'.$dateNumericZone;
            }
        }
        if (date_default_timezone_set($dateNumericZone)) {
            error_reporting($errorReportingBefore);
            return;
        }

        $defaultZone = 'UTC';
        if (date_default_timezone_set($defaultZone)) {
            error_reporting($errorReportingBefore);
            return;
        }
        Utils::logError(__METHOD__,
            sprintf("Failed to set timezone (all are invalid: '%s', '%s', '%s', '%s')",
                $timedatectlZone, $dateZone, $dateNumericZone, $defaultZone));

        error_reporting($errorReportingBefore);
    }

    /**
     * Сделать запись в лог модуля Itgrix
     *
     * Путь к логу прописан в settings.json.
     *
     * @param string $action  Название команды (будет записано в поле 'command')
     * @param array  $params  Параметры кастомизации (ассоциативный массив 'поле' => 'значение')
     * @param string $level   Уровень логирования (поддерживаются: 'debug', 'info', 'warning', 'error')
     * @param string $message Сообщение
     *
     * @throws Exception В случае ошибки
     */
    public static function log($action, array $params, $level, $message) {

        // Специальное получение миллисекунд
        // для совместимости со старыми версиями PHP
        $microtime = explode(' ', microtime());
        $milliseconds = intval(floatval($microtime[0]) * 1000);
        $timestamp = intval($microtime[1]);

        $date = new DateTime('now');
        date_timestamp_set($date, $timestamp);

        $logEntry = array(
            '0_ts'         => sprintf('%s.%s %s',
                $date->format('Y-m-d H:i:s'),
                str_pad($milliseconds, 3, '0', STR_PAD_LEFT),
                $date->format('P')),
            '1_level'      => $level,
            '3_call_stage' => 'PHP customizer',
            '4_command'    => $action,
            '5_msg'        => $message,
        );

        $id = self::getCallId($params);
        if ($id != '') {
            $logEntry['2_call_id'] = $id;
        }

        ksort($logEntry);

        $logEntrySerialized = json_encode($logEntry, JSON_UNESCAPED_UNICODE);

        if ((preg_match('/"1_level":"([a-z]+)"/', $logEntrySerialized, $matches) !== 1) ||
            (count($matches) !== 2)) {

            throw new Exception(sprintf(
                "failed to find correct 1_level in serialized log entry: '%s'",
                $logEntrySerialized));
        }
        $padding = str_pad('', MAX_LOG_LEVEL_NAME_LENGTH - strlen($matches[1]));

        $logEntryFinal = preg_replace('/("1_level":"[a-z]+"),/', '$1,'.$padding,
            $logEntrySerialized, 1);
        if (($logEntryFinal === null) ||
            (strlen($logEntryFinal) != strlen($logEntrySerialized) + strlen($padding))) {

            throw new Exception(sprintf(
                "failed to add padding to 1_level in serialized log entry: '%s'",
                $logEntrySerialized));
        }
        file_put_contents(Settings::getLogPath(), $logEntryFinal."\n", FILE_APPEND);
    }

    /**
     * Сделать запись уровня debug в лог
     *
     * Путь к логу прописан в settings.json.
     *
     * Данная функция предназначена для служебного использования.
     * В клиентских кастомизациях нужно использовать функцию экземпляра класса Customizer.
     *
     * @param string $functionName Имя функции, вызывающей логирование
     *                             (будет записано в поле 'command')
     * @param string $message      Сообщение
     *
     * @throws Exception В случае ошибки
     */
    public static function logDebug($functionName, $message) {
        self::log($functionName, array(), 'debug', $message);
    }

    /**
     * Сделать запись уровня info в лог
     *
     * Путь к логу прописан в settings.json.
     *
     * Данная функция предназначена для служебного использования.
     * В клиентских кастомизациях нужно использовать функцию экземпляра класса Customizer.
     *
     * @param string $functionName Имя функции, вызывающей логирование
     *                             (будет записано в поле 'command')
     * @param string $message      Сообщение
     *
     * @throws Exception В случае ошибки
     */
    public static function logInfo($functionName, $message) {
        self::log($functionName, array(), 'info', $message);
    }

    /**
     * Сделать запись уровня warning в лог
     *
     * Путь к логу прописан в settings.json.
     *
     * Данная функция предназначена для служебного использования.
     * В клиентских кастомизациях нужно использовать функцию экземпляра класса Customizer.
     *
     * @param string $functionName Имя функции, вызывающей логирование
     *                             (будет записано в поле 'command')
     * @param string $message      Сообщение
     *
     * @throws Exception В случае ошибки
     */
    public static function logWarn($functionName, $message) {
        self::log($functionName, array(), 'warning', $message);
    }

    /**
     * Сделать запись уровня error в лог
     *
     * Путь к логу прописан в settings.json.
     *
     * Данная функция предназначена для служебного использования.
     * В клиентских кастомизациях нужно использовать функцию экземпляра класса Customizer.
     *
     * @param string $functionName Имя функции, вызывающей логирование
     *                             (будет записано в поле 'command')
     * @param string $message      Сообщение
     *
     * @throws Exception В случае ошибки
     */
    public static function logError($functionName, $message) {
        self::log($functionName, array(), 'error', $message);
    }

    /**
     * Прочитать JSON из файла
     *
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Путь к логу прописан в settings.json.
     *
     * @param string $fileName Имя файла
     *
     * @return bool|mixed Данные JSON в виде ассоциативного массива,
     *                    либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function getJson($fileName) {

        $contents = file_get_contents($fileName);
        if ($contents === false) {
            self::logError(__METHOD__,
                "Failed to read file: '$fileName'");
            return false;
        }

        $json = json_decode($contents, true);
        if ($json === null) {
            self::logError(__METHOD__,
                "Failed to parse JSON file: '$fileName'");
            return false;
        }
        return $json;
    }

    /**
     * Получить данные из конфига
     *
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Пути к конфигу и логу прописаны в settings.json.
     *
     * @return bool|mixed Данные JSON-конфига в виде ассоциативного массива,
     *                    либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function getConfig() {
        return self::getJson(Settings::getConfigPath());
    }

    /**
     * Сделать запрос в MySQL-базу с помощью php-mysqli
     *
     * Реквизиты для доступа в БД читаются из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Пути к конфигу и логу прописаны в settings.json.
     *
     * Если по каким-то причинам невозможно воспользоваться php-mysqli,
     * можно воспользоваться функцией mysqlClcQuery,
     * которая делает то же самое, но с помощью MySQL Command-Line Client.
     *
     * @param string $sql       Строка с SQL-запросом
     * @param bool   $getResult Нужно ли получать результат запроса.
     *                          Например, для SELECT-запросов имеет смысл получать,
     *                          а для UPDATE-запросов нет.
     *
     * @return array|bool Массив с результом запроса (массив ассоциативных массивов 'поле' => 'значение'),
     *                    либо true в случае успеха без получения результата,
     *                    либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function mysqliQuery($sql, $getResult = false) {

        if (! function_exists('mysqli_query')) {
            self::logError(__METHOD__,
                "Function 'mysqli_query' does not exist.".
                " Check if php-mysqli is installed on your system.".
                " Or use Utils::mysqlClcQuery instead.");
            return false;
        }

        $config = self::getConfig();
        if ($config === false) {
            self::logError(__METHOD__,
                'Failed to get config');
            return false;
        }

        $config = $config['database'];

        $dbConn = mysqli_connect(
            $config['host'],
            $config['login'],
            $config['password'],
            $config['database'],
            $config['port']
        );

        if (! $dbConn) {
            $err = mysqli_connect_error();
            self::logError(__METHOD__,
                "Failed to open connection to MySQL server: '$err'");
            return false;
        }

        $dbResult = mysqli_query($dbConn, $sql);
        if ($dbResult === false) {
            $err = mysqli_error($dbConn);
            self::logError(__METHOD__,
                "MySQL query failed: '$sql'. Error: '$err'");
            return false;
        }

        if (! $getResult) {
            return true;
        }

        $result = array();
        while ($row = mysqli_fetch_assoc($dbResult)) {
            $result[] = $row;
        }
        return $result;
    }

    /**
     * Сделать запрос в MySQL-базу с помощью MySQL Command-Line Client
     *
     * Реквизиты для доступа в БД читаются из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Пути к конфигу и логу прописаны в settings.json.
     *
     * Данной функцией имеет смысл пользоваться, если по каким-то причинам
     * невозможно воспользоваться функцией mysqliQuery,
     * которая делает то же самое, но с помощью php-mysqli.
     *
     * @param string $sql       Строка с SQL-запросом
     * @param bool   $getResult Нужно ли получать результат запроса.
     *                          Например, для SELECT-запросов имеет смысл получать,
     *                          а для UPDATE-запросов нет.
     *
     * @return array|bool Массив с результом запроса (массив ассоциативных массивов 'поле' => 'значение'),
     *                    либо true в случае успеха без получения результата,
     *                    либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function mysqlClcQuery($sql, $getResult = false) {

        $config = self::getConfig();
        if ($config === false) {
            self::logError(__METHOD__,
                'Failed to get config');
            return false;
        }

        $config = $config['database'];

        if (empty($config['password'])) {
            $passwordArg = '';
        } else {
            $passwordArg = '-p' . preg_replace(
                    '~([:#;.\\/!*?$@()\[\]><&|{}])~',
                    '\\$1',
                    $config['password']);
        }

        $cmd = sprintf('mysql -u%s %s -h%s --port=%d %s -e "%s" 2>&1',
            $config['login'],
            $passwordArg,
            $config['host'],
            $config['port'],
            $config['database'],
            str_replace('"', '\"', $sql));

        exec($cmd, $res, $exitCode);

        if ($exitCode !== 0) {
            self::logError(__METHOD__,
                "mysql command exited with code '$exitCode' and output[0]: '$res[0]'");
            return false;
        }

        if (! $getResult) {
            return true;
        }

        $result = array();

        if (count($res) > 1) {
            $columns = explode("\t", array_shift($res));

            foreach ($res as $row) {
                $e = explode("\t", $row);
                $_ = array();

                foreach ($columns as $i => $key) {
                    $_[$key] = $e[$i];
                }

                $result[] = $_;
                unset($_, $e, $row);
            }
        }
        return $result;
    }

    /**
     * Конвертировать заданный файл в mp3
     *
     * Исходный файл НЕ удаляется.
     * Если файл уже имеет расширение mp3, то ничего не происходит.
     *
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Путь к логу прописан в settings.json.
     *
     * @param string $recPath Путь к исходному файлу записи
     *
     * @return bool|string Путь к mp3-файлу,
     *                     либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function convertToMp3($recPath) {
        if (! file_exists($recPath)) {
            self::logError(__METHOD__,
                'Recording file does not exist: '.var_export($recPath, true));
            return false;
        }
        $pathInfo = pathinfo($recPath);
        $recDirname   = $pathInfo['dirname'];
        $recExtension = $pathInfo['extension'];
        $recFilename  = $pathInfo['filename'];

        if (strtolower($recExtension) == 'mp3') {
            self::logWarn(__METHOD__,
                'Recording file is already in mp3 format');
            return $recPath;
        }

        if (strtolower($recExtension) != 'wav') {
            self::logError(__METHOD__,
                "Recording file extension is not supported: '$recExtension'");
            return false;
        }

        $newPath = "{$recDirname}/{$recFilename}.mp3";

        $command = "lame -h -b 192 '{$recPath}' '{$newPath}' > /dev/null 2>&1";
        exec($command, $_, $exitCode);

        if ($exitCode === 127) {
            self::logError(__METHOD__,
                'Failed to run LAME. '.
                'Check if it is installed on your system and has permission to execute');
            return false;
        } else if ($exitCode !== 0) {
            self::logError(__METHOD__,
                "LAME failed with exit code: $exitCode");
            return false;
        }

        return $newPath;
    }

    /**
     * Сделать GET-запрос с помощью php-curl
     *
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Путь к логу прописан в settings.json.
     *
     * @param string $url         URL-адрес для запроса
     * @param array  $params      Параметры запроса, добавляются в URL (ассоциативный массив 'параметр' => 'значение')
     * @param array  $options     Дополнительные опции php-curl (ассоциативный массив 'CURLOPT_*' => 'значение').
     *                            Заданные опции добавляются к опциям по умолчанию (см. тело функции).
     *                            Полный список опций доступен в документации:
     *                            https://www.php.net/manual/ru/function.curl-setopt.php
     *
     * @return array Массив с данными о выполнении запроса.
     *               Поля описаны в справке функции Utils::execCurlRequest.
     *
     * @throws Exception В случае ошибки
     */
    public static function curlGet($url, array $params = array(), array $options = array()) {

        $defaultOptions = array(
            CURLOPT_URL            => $url.'?'.http_build_query($params),
            CURLOPT_HEADER         => false,
            CURLOPT_TIMEOUT        => Settings::getRequestTimeoutSec(),
        );
        return self::execCurlRequest(self::makeCurlOptions($options, $defaultOptions));
    }

    /**
     * Сделать POST-запрос с помощью php-curl
     *
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Путь к логу прописан в settings.json.
     *
     * @param string $url         URL-адрес для запроса
     * @param string $contentType MIME-тип содержимого. По умолчанию 'text/plain'.
     * @param string $body        Тело запроса. По умолчанию пустая строка.
     * @param array  $options     Дополнительные опции php-curl (ассоциативный массив 'CURLOPT_*' => 'значение').
     *                            Заданные опции добавляются к опциям по умолчанию (см. тело функции).
     *                            Полный список опций доступен в документации:
     *                            https://www.php.net/manual/ru/function.curl-setopt.php
     *
     * @return array Массив с данными о выполнении запроса.
     *               Поля описаны в справке функции Utils::execCurlRequest.
     *
     * @throws Exception В случае ошибки
     */
    public static function curlPost($url, $contentType = 'text/plain', $body = '', array $options = array()) {

        $defaultOptions = array(
            CURLOPT_URL            => $url,
            CURLOPT_POST           => true,
            CURLOPT_HEADER         => false,
            CURLOPT_TIMEOUT        => Settings::getRequestTimeoutSec(),
            CURLOPT_POSTFIELDS     => $body,
            CURLOPT_HTTPHEADER     => array(
                'Content-Type: '.$contentType,
            ),
        );
        return self::execCurlRequest(self::makeCurlOptions($options, $defaultOptions));
    }

    /**
     * Сделать PATCH-запрос с помощью php-curl
     *
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Путь к логу прописан в settings.json.
     *
     * @param string $url         URL-адрес для запроса
     * @param string $contentType MIME-тип содержимого. По умолчанию 'text/plain'.
     * @param string $body        Тело запроса. По умолчанию пустая строка.
     * @param array  $options     Дополнительные опции php-curl (ассоциативный массив 'CURLOPT_*' => 'значение').
     *                            Заданные опции добавляются к опциям по умолчанию (см. тело функции).
     *                            Полный список опций доступен в документации:
     *                            https://www.php.net/manual/ru/function.curl-setopt.php
     *
     * @return array Массив с данными о выполнении запроса.
     *               Поля описаны в справке функции Utils::execCurlRequest.
     *
     * @throws Exception В случае ошибки
     */
    public static function curlPatch($url, $contentType = 'text/plain', $body = '', array $options = array()) {

        $defaultOptions = array(
            CURLOPT_URL            => $url,
            CURLOPT_CUSTOMREQUEST  => 'PATCH',
            CURLOPT_HEADER         => false,
            CURLOPT_TIMEOUT        => Settings::getRequestTimeoutSec(),
            CURLOPT_POSTFIELDS     => $body,
            CURLOPT_HTTPHEADER     => array(
                'Content-Type: '.$contentType,
            ),
        );
        return self::execCurlRequest(self::makeCurlOptions($options, $defaultOptions));
    }

    /**
     * Разбор URL (обёртка для parse_url с исправлением)
     *
     * @param string $url URL для разбора
     *
     * @return bool|array Ассоциативный массив с компонентами URL,
     *                    либо false в случае ошибки.
     *                    Подробное описание:
     *                    https://www.php.net/manual/ru/function.parse-url.php
     */
    public static function parseUrl($url) {

        // parse_url парсит адрес, состоящий только из хоста как path,
        // поэтому, если схемы нет, добавим её, распарсим и удалим из результата
        $addedScheme = false;
        // Спецификация scheme:
        // https://datatracker.ietf.org/doc/html/rfc1738#section-2.1
        // https://datatracker.ietf.org/doc/html/rfc3986#section-3.1
        $schemedRe = '/^([a-zA-Z][a-zA-Z\d\+\-\.]*\:\/\/).*/';
        if (preg_match($schemedRe, $url, $matches) !== 1) {
            $addedScheme = true;
            $url = 'http://'.$url;
        }

        $parsedUrl = parse_url($url);

        if ($parsedUrl === false) {
            return false;
        }

        if ($addedScheme) {
            $parsedUrl['scheme'] = NULL;
        }
        return $parsedUrl;
    }

    /**
     * Сделать запрос на HTTP-сервер модуля Itgrix
     *
     * Все необходимые для запроса реквизиты будут получены из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Путь к логу прописан в settings.json.
     *
     * @param string $node       Узел API модуля Itgrix
     * @param string $httpMethod HTTP-метод запроса
     * @param array  $params     Параметры запроса в виде ассоциативного массива
     *
     * @return array|false Массив с данными HTTP-ответа:
     *                       'httpCode'  => (int)    Код ответа HTTP;
     *                       'response'  => (string) Ответ на запрос (тело).
     *                     Либо false в случае ошибки выполнения запроса.
     *
     * @throws Exception В случае ошибки
     */
    public static function doRequestToModule($node, $httpMethod, array $params) {

        $config = self::getConfig();
        if ($config === false) {
            self::logError(__METHOD__,
                'Failed to get config');
            return false;
        }
        $config = $config['server'];

        $login    = $config['login'];
        $password = $config['password'];

        $url = self::makeModuleUrl() . '/' . $node;

        $curlOptions = array(
            CURLOPT_HTTPAUTH       => CURLAUTH_DIGEST,
            CURLOPT_USERPWD        => "$login:$password",
            CURLOPT_SSL_VERIFYHOST => 0,
            CURLOPT_SSL_VERIFYPEER => false,
            CURLOPT_FOLLOWLOCATION => true,
        );

        // Поля $result описаны в справке функции Utils::execCurlRequest
        if ($httpMethod == 'GET') {
            $result = Utils::curlGet($url, $params, $curlOptions);
        } else if ($httpMethod == 'POST') {
            $result = Utils::curlPost($url,
                'application/json', json_encode($params), $curlOptions);
        } else {
            self::logError(__METHOD__,
                sprintf("Unsupported HTTP method: '%s'", $httpMethod));
            return false;
        }

        if ($result['success'] === false) {
            self::logError(__METHOD__, sprintf(
                "Request failed. " .
                "Node: '%s'; HTTP method: '%s'; params: '%s'; " .
                "errorInfo: '%s'; httpCode: '%s'; response: '%s'",
                $node, $httpMethod, json_encode($params, JSON_UNESCAPED_UNICODE),
                json_encode($result['errorInfo'], JSON_UNESCAPED_UNICODE),
                $result['httpCode'], $result['response']));
            return false;
        }

        return array(
            'httpCode' => $result['httpCode'],
            'response' => $result['response'],
        );
    }

    /**
     * Приготовить запрос для отправки в CRM
     * (для передачи в функцию doRequestArrayToCrmViaModule)
     *
     * @param string $crmApiMethod Название метода в API CRM (допускаются слэши - '/')
     * @param string $httpMethod   HTTP-метод
     * @param array  $params       Параметры запроса в виде ассоциативного массива
     *
     * @return array Ассоциативный массив c данными запроса
     *               для отправки в функцию doRequestArrayToCrmViaModule
     */
    public static function prepareRequestToCrm($crmApiMethod, $httpMethod, $params) {
        return array(
            'crmApiMethod'   => $crmApiMethod,
            'httpMethod'     => $httpMethod,
            'params'         => $params,
        );
    }

    /**
     * Сделать несколько запросов в CRM через модуль Itgrix
     *
     * Все необходимые для запроса реквизиты будут получены из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Путь к логу прописан в settings.json.
     *
     * @param array $requestArray Массив запросов,
     *   каждый из которых должен быть ассоциативным массивом:
     *     'crmApiMethod' => Название метода в API CRM (допускаются слэши - '/')
     *     'httpMethod'   => HTTP-метод
     *     'params'       => Параметры запроса в виде ассоциативного массива
     *
     * @return array|false Массив с данными для каждого запроса
     *   в виде ассоциативного массива:
     *     'success' => (bool)  Успех выполения запроса модулем Itgrix;
     *     'result'  => (mixed) Результат запроса:
     *         либо (в случае успеха) данные от CRM (извлеченные модулем Itgrix из ответа CRM),
     *         либо (если НЕ успех) JSON-объект с описанием ошибки в поле error.
     *   Либо false в случае ошибки выполнения запроса через модуль Itgrix.
     *
     * @throws Exception В случае ошибки
     */
    public static function doRequestArrayToCrmViaModule(array $requestArray) {
        $requests = array();
        $crmApiMethods = array();
        foreach ($requestArray as $request) {
            array_push($requests, array(
                'node'   => $request['crmApiMethod'],
                'method' => $request['httpMethod'],
                'params' => $request['params'],
            ));
            array_push($crmApiMethods, $request['crmApiMethod']);
        }

        $result = self::doRequestToModule('crm/request', 'POST',
            array('requests' => $requests));
        if ($result === false) {
            self::logError(__METHOD__, sprintf(
                "Failed to make multiple requests to CRM via module. " .
                    "CRM API methods: %s",
                json_encode($crmApiMethods)));
            return false;
        }
        $httpCode = $result['httpCode'];
        $response = $result['response'];

        // Штатно модуль возвращает 200 OK или 424 Failed Dependency
        // Если хотя бы один запрос закончился ошибкой, модуль вернет 424
        if (! in_array($httpCode, array(200, 424), true)) {
            self::logError(__METHOD__, sprintf(
                "Failed to make multiple requests to CRM via module: " .
                    "unexpected HTTP code: '%s'. " .
                    "CRM API methods: %s; response: '%s'",
                $httpCode, json_encode($crmApiMethods), $response));
            return false;
        }

        $responseData = json_decode($response, true);
        if ($responseData === null) {
            Utils::logError(__METHOD__, sprintf(
                "Failed to parse JSON response: '%s'", $response));
            return false;
        }
        if (! isset($responseData['results'])) {
            Utils::logError(__METHOD__, sprintf(
                "No 'results' field in response: '%s'", $response));
            return false;
        }
        if (gettype($responseData['results']) !== 'array') {
            Utils::logError(__METHOD__, sprintf(
                "'results' field in response is not an array: '%s'", $response));
            return false;
        }
        if (count($responseData['results']) !== count($requests)) {
            Utils::logError(__METHOD__, sprintf(
                "'results' array length (%d) differs from request count (%d). Response: '%s'",
                count($responseData['results']), count($requests), $response));
            return false;
        }

        $results = array();
        foreach ($responseData['results'] as $requestResult) {
            $success = true;
            if ((gettype($requestResult) === 'array') && isset($requestResult['error'])) {
                $success = false;
            }
            array_push($results, array(
                'success' => $success,
                'result'  => $requestResult,
            ));
        }
        return $results;
    }

    /**
     * Сделать запрос в CRM через модуль Itgrix
     *
     * Все необходимые для запроса реквизиты будут получены из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Путь к логу прописан в settings.json.
     *
     * @param string $crmApiMethod Название метода в API CRM (допускаются слэши - '/')
     * @param string $httpMethod   HTTP-метод
     * @param array  $params       Параметры запроса в виде ассоциативного массива
     *
     * @return array|false Ассоциативный массив:
     *     'success' => (bool)  Успех выполения запроса модулем Itgrix;
     *     'result'  => (mixed) Результат запроса:
     *         либо (в случае успеха) данные от CRM (извлеченные модулем Itgrix из ответа CRM),
     *         либо (если НЕ успех) JSON-объект с описанием ошибки в поле error.
     *   Либо false в случае ошибки выполнения запроса через модуль Itgrix.
     *
     * @throws Exception В случае ошибки
     */
    public static function doRequestToCrmViaModule($crmApiMethod, $httpMethod, $params) {
        $results = self::doRequestArrayToCrmViaModule(array(
            self::prepareRequestToCrm($crmApiMethod, $httpMethod, $params)));
        if ($results === false) {
            self::logError(__METHOD__, sprintf(
                "Failed to make a request to CRM via module. " .
                "CRM API method: '%s'; HTTP method: '%s'; params: '%s'",
                $crmApiMethod, $httpMethod, json_encode($params, JSON_UNESCAPED_UNICODE)));
            return false;
        }
        return $results[0];
    }

    /**
     * Сделать URL-адрес модуля Itgrix
     *
     * Адрес модуля формируется исходя из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Пути к конфигу и логу прописаны в settings.json.
     *
     * @param string $address Использовать заданный адрес вместо стандартного адреса сервера.
     *                        В стандартном адресе схема, хост, порт будут заменены
     *                        на соответствующие элементы заданного адреса, если они присутствуют.
     *                        Например, можно передать поле конфига <CRM>.recording_address,
     *                        чтобы построить адрес модуля для публикации ссылки на записи
     *                        разговоров в CRM.
     *
     * @return bool|string Строка, содержащая URL,
     *                     либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function makeModuleUrl($address = '') {

        $config = self::getConfig();
        if ($config === false) {
            self::logError(__METHOD__,
                'Failed to get config');
            return false;
        }

        $config = $config['server'];
        $url = array();

        $url['scheme'] = 'http';
        $url['port'] = $config['port'];

        if ($config['use_https'] === true) {
            $url['scheme'] = 'https';
            $url['port'] = $config['secure_port'];
        }

        $url['host'] = 'localhost';

        if (($config['host'] !== '') && ($config['host'] !== '0.0.0.0')) {
            $url['host'] = $config['host'];
        }

        if ($address !== '') {
            $parsedUrl = self::parseUrl($address);
            if ($parsedUrl === false) {
                self::logError(__METHOD__,
                    "Failed to parse address: '$address'");
                return false;
            }
            foreach ($url as $k => $v) {
                if (isset($parsedUrl[$k])) {
                    $url[$k] = $parsedUrl[$k];
                }
            }
        }

        return sprintf('%s://%s:%s',
            $url['scheme'], $url['host'], $url['port']);
    }

    /**
     * Сделать URL-адрес для получения записи разговора
     *
     * Адрес модуля формируется исходя из конфига
     * (поле <CRM>.recording_address, а также server.*).
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Пути к конфигу и логу, а также поле конфига <CRM> прописаны в settings.json.
     *
     * @param string $id   ID разговора
     * @param string $from Номер звонящего
     * @param string $to   Номер, на который пришёл звонок
     *
     * @return bool|string Строка, содержащая URL,
     *                     либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function makeRecordingUrl($id, $from, $to) {

        $config = Utils::getConfig();
        if ($config === false) {
            Utils::logError(__METHOD__,
                'Failed to get config');
            return false;
        }
        $recordingAddress = '';
        if (isset($config[Settings::getConfigCrmField()]['recording_address'])) {
            $recordingAddress = $config[Settings::getConfigCrmField()]['recording_address'];
        }

        $moduleAddress = self::makeModuleUrl($recordingAddress);

        if ($moduleAddress === false) {
            return false;
        }
        return sprintf('%s/get_conversation_recording?id=%s&from=%s&to=%s',
            $moduleAddress, $id, $from, $to);
    }

    /**
     * Запросить у модуля токен доступа в CRM
     *
     * Адрес модуля формируется исходя из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Пути к конфигу и логу прописаны в settings.json.
     *
     * @return bool|string Строка, содержащая токен доступа (access_token),
     *                     либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function getAccessToken() {

        $config = self::getConfig();
        if ($config === false) {
            self::logError(__METHOD__,
                'Failed to get config');
            return false;
        }

        $config = $config['server'];

        $login    = $config['login'];
        $password = $config['password'];

        $url = self::makeModuleUrl().'/crm/access_token';

        // Поля $result описаны в справке функции Utils::execCurlRequest
        $result = self::curlGet($url, array(), array(
            CURLOPT_HTTPAUTH => CURLAUTH_DIGEST,
            CURLOPT_USERPWD  => "$login:$password",
            CURLOPT_SSL_VERIFYHOST => 0,
            CURLOPT_SSL_VERIFYPEER => false,
        ));

        if (($result['success'] === false) || ($result['httpCode'] !== 200)) {
            self::logError(__METHOD__, sprintf(
                'Request failed (code: %s)',
                $result['httpCode']));
            return false;
        }
        return $result['response'];
    }

    /**
     * Запросить у модуля поставить запрос к CRM в общую очередь запросов
     * с учётом ограничений на нагрузку API CRM
     * и дождаться разрешения на его выполнение.
     *
     * Перед запросом в CRM сначала НЕОБХОДИМО сделать запрос ratelimit в модуль.
     *
     * Адрес модуля формируется исходя из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Пути к конфигу и логу прописаны в settings.json.
     *
     * @return bool Успех запроса
     *
     * @throws Exception В случае ошибки
     */
    public static function waitCrmApiRatelimit() {

        $config = self::getConfig();
        if ($config === false) {
            self::logError(__METHOD__,
                'Failed to get config');
            return false;
        }

        $config = $config['server'];

        $login    = $config['login'];
        $password = $config['password'];

        $url = self::makeModuleUrl().'/crm/ratelimit';

        // Поля $result описаны в справке функции Utils::execCurlRequest
        $result = self::curlGet($url, array(), array(
            CURLOPT_HTTPAUTH => CURLAUTH_DIGEST,
            CURLOPT_USERPWD  => "$login:$password",
            CURLOPT_SSL_VERIFYHOST => 0,
            CURLOPT_SSL_VERIFYPEER => false,
        ));

        if (($result['success'] === false) || ($result['httpCode'] !== 200)) {
            self::logError(__METHOD__, sprintf(
                'Request failed (code: %s)',
                $result['httpCode']));
            return false;
        }
        return true;
    }

//==============================================================================

    protected static function getCallId(array $params) {

        if (isset($params['call_full']['channel']['id'])) {
            return $params['call_full']['channel']['id'];
        }
        else if (isset($params['parent_channel_id'])) {
            return $params['parent_channel_id'];
        }
        else if (isset($params['id'])) {
            return $params['id'];
        }
        else if (isset($params['call']['id'])) {
            return $params['call']['id'];
        }
        else if (isset($params['conversation']['parent_channel_id'])) {
            return $params['conversation']['parent_channel_id'];
        }
        else {
            return '';
        }
    }

    protected static function curlOptionsToNamed(array $options) {

        $definedConsts = get_defined_constants(true);
        $curlConsts = $definedConsts['curl'];

        $curloptConsts = array();
        foreach ($curlConsts as $name => $value) {
            if (strpos($name, 'CURLOPT_') === 0) {
                $curloptConsts[$name] = $value;
            }
        }
        $curloptNames = array_flip($curloptConsts);

        $optionsNamed = array();

        foreach (array_keys($options) as $optCode) {
            $optName = 'undefined';
            if (isset($curloptNames[$optCode])) {
                $optName = $curloptNames[$optCode];
            }
            $optionsNamed[$optName] = $options[$optCode];
        }
        return $optionsNamed;
    }

    protected static function makeCurlOptions($options, $defaultOptions) {

        $curlOpts = $options + $defaultOptions;

        foreach (array_keys($defaultOptions) as $opt) {
            if ((gettype($defaultOptions[$opt]) === 'array') &&
                (isset($options[$opt]))) {

                $curlOpts[$opt] = array_merge($options[$opt], $defaultOptions[$opt]);
            }
        }
        return $curlOpts;
    }

    protected static function curlOptionsAuthCensor(array $options) {

        if (isset($options[CURLOPT_HTTPHEADER])) {
            foreach ($options[CURLOPT_HTTPHEADER] as &$header) {
                if (preg_match('/\ *Authorization\:.*/', $header)) {

                    $res = preg_replace(
                        '/\ *(Authorization\:)\ *([^\ ]+)\ *[^\ ]+([^\ ]{8})/',
                        '$1 $2 *$3',
                        $header);
                    if ($res !== null) {
                        $header = $res;
                    }

                    return $options;
                }
            }
        }

        if (isset($options[CURLOPT_POSTFIELDS])) {
            $res = preg_replace(
                '/auth=[a-zA-Z0-9]+([a-zA-Z0-9]{8})/',
                'auth=*$1',
                $options[CURLOPT_POSTFIELDS]);
            if ($res !== null) {
                $options[CURLOPT_POSTFIELDS] = $res;
            }
        }

        return $options;
    }

    /**
     * Выполнить запрос с помощью curl (должен быть установлен php-curl)
     *
     * @param array $options           Список опций для curl.
     *                                 Примечание: опция CURLOPT_RETURNTRANSFER
     *                                 в любом случае будет установлена true.
     *
     * @return array                   Массив с данными о выполнении запроса:
     *   'success'   => (bool)         Успешность выполнения curl_exec.
     *                                 Примечание: будет true, даже если код возврата HTTP не 2xx
     *   'errorInfo' => (array)        Массив с данными об ошибке curl.
     *                                 Пустой, если до запроса (curl_exec) даже не дошло.
     *     'curlErrno' => curl_errno
     *     'curlError' => curl_error
     *     'curlInfo'  => curl_getinfo
     *   'httpCode'  => (int)          Код ответа HTTP. Равен 0, если код не получен.
     *   'response'  => (string|null)  Ответ на запрос (тело),
     *                                 либо null в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    protected static function execCurlRequest(array $options) {

        if (! function_exists('curl_exec')) {
            self::logError(__METHOD__,
                "Function 'curl_exec' does not exist.".
                " Check if php-curl is installed on your system.");
            return array(
                'success'   => false,
                'errorInfo' => array(),
                'httpCode'  => 0,
                'response'  => null,
            );
        }

        $curl = curl_init();
        if ($curl === false) {
            self::logError(__METHOD__,
                "Failed to init curl");
            return array(
                'success'   => false,
                'errorInfo' => array(),
                'httpCode'  => 0,
                'response'  => null,
            );
        }

        $options[CURLOPT_RETURNTRANSFER] = true;
        $optionsEncoded = json_encode(
            self::curlOptionsToNamed(self::curlOptionsAuthCensor($options)));

        if (curl_setopt_array($curl, $options) === false) {
            self::logError(__METHOD__,
                "Failed set curl options: $optionsEncoded");
            return array(
                'success'   => false,
                'errorInfo' => array(),
                'httpCode'  => 0,
                'response'  => null,
            );
        }

//        curl_setopt($curl, CURLOPT_VERBOSE, true);
        $execResult = curl_exec($curl);

        $errorInfo = array(
            'curlErrno' => curl_errno($curl),
            'curlError' => curl_error($curl),
            'curlInfo'  => curl_getinfo($curl),
        );
        $httpCode = 0;
        if (isset($errorInfo['curlInfo']['http_code'])) {
            $httpCode = $errorInfo['curlInfo']['http_code'];
        }

        curl_close($curl);

        if ($execResult === false) {

            self::logError(__METHOD__, sprintf(
                "Curl exec failed. Error info: %s. Options: %s",
                json_encode($errorInfo),
                $optionsEncoded));

            return array(
                'success'   => false,
                'errorInfo' => $errorInfo,
                'httpCode'  => $httpCode,
                'response'  => null,
            );
        }

        if (($httpCode < 200) || ($httpCode >= 300)) {
            self::logWarn(__METHOD__,
                "Response status is NOT 2xx: $httpCode;".
                " request options: $optionsEncoded;".
                " errorInfo: ".json_encode($errorInfo).';'.
                " response: '$execResult'");
        }

        return array(
            'success'   => true,
            'errorInfo' => $errorInfo,
            'httpCode'  => $httpCode,
            'response'  => $execResult,
        );
    }

}
