<?php

require_once __DIR__.'/Settings.php';
require_once __DIR__.'/Utils.php';

/**
 * Class Bitrix
 *
 * Реализует пользовательский API кастомизатора для запросов в CRM Битрикс24.
 *
 * Состоит из произвольного запроса в CRM и реализации конкретных запросов
 * (например, создание задачи).
 * По необходимости список конкретных запросов будет расширяться.
 *
 * Все методы статические (здесь НЕ используется экземляр Customizer).
 */
class Bitrix {

    /**
     * Произвольный запрос в REST API Битрикс24
     *
     * Сначала будет отправлен запрос в модуль Itgrix на получение токена доступа в Битрикс.
     *
     * Все необходимые для запросов реквизиты будут получены из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Пути к конфигу и логу прописаны в settings.json
     *
     * @param string $restMethod Название метода REST API
     * @param string $httpMethod HTTP-метод. Поддерживаются: 'GET', 'POST'
     * @param array  $data       Данные запроса (ассоциативный массив 'поле' => 'значение')
     *
     * @return array|bool Массив с данными HTTP-ответа (
     *                      'httpCode' - код ответа,
     *                      'response' - тело ответа),
     *                    либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function request($restMethod, $httpMethod, array $data = array()) {
        // HTTP Method

        $httpMethod = strtoupper($httpMethod);
        if (! in_array($httpMethod, array('GET', 'POST'))) {
            Utils::logError(__METHOD__,
                "HTTP method is not supported: '$httpMethod'");
            return false;
        }

        // Token

        $accessToken = Utils::getAccessToken();
        if ($accessToken === false) {
            Utils::logError(__METHOD__,
                'Failed to get access token');
            return false;
        }

        // Portal

        $config = Utils::getConfig();
        if ($config === false) {
            Utils::logError(__METHOD__,
                'Failed to get config: '.Settings::getConfigPath());
            return false;
        }
        if ((! isset($config['bitrix']['host'])) || (trim($config['bitrix']['host']) == '')) {
            Utils::logError(__METHOD__,
                "Config field does not exist or is empty: 'bitrix.host'");
            return false;
        }
        $portal = rtrim($config['bitrix']['host'], '/');

        // Prepare data

        $url = "$portal/rest/$restMethod";
        $data['auth'] = $accessToken;

        // Rate limit request

        $ratelimitResult = Utils::waitCrmApiRatelimit();
        if ($ratelimitResult === false) {
            Utils::logError(__METHOD__,
                'ratelimit request failed');
            return false;
        }

        // CRM request

        $result = false;
        if ($httpMethod == 'GET') {
            $result = Utils::curlGet($url, $data);
        } else if ($httpMethod == 'POST') {
            $result = Utils::curlPost($url,
                'application/x-www-form-urlencoded', http_build_query($data));
        }
        // Поля $result описаны в справке функции Utils::execCurlRequest

        if ($result === false) {
            Utils::logError(__METHOD__, sprintf(
                "Bitrix request failed: %s. REST method: %s. HTTP method: %s",
                "no result",
                $restMethod,
                $httpMethod));
            return false;
        } else if ($result['success'] === false) {
            Utils::logError(__METHOD__, sprintf(
                "Bitrix request failed: %s. REST method: %s. HTTP method: %s",
                "execution error",
                $restMethod,
                $httpMethod));
            return false;
        } else if ($result['httpCode'] !== 200) {
            // Bitrix REST API returns JSON
            // php-xml or php-mbstring packages may be missing
            $unescapedResponse = json_encode(json_decode(
                $result['response']), JSON_UNESCAPED_UNICODE);

            Utils::logError(__METHOD__, sprintf(
                "Bitrix request failed: ".
                    "%s. REST method: %s. HTTP method: %s. ".
                    "Response: %s. Response with unescaped unicode: %s",
                "HTTP code is not 200: ".$result['httpCode'],
                $restMethod,
                $httpMethod,
                $result['response'],
                $unescapedResponse));
            return false;
        }
        return array(
            'httpCode' => $result['httpCode'],
            'response' => $result['response'],
        );
    }

    /**
     * Создать задачу в Битрикс
     *
     * Все необходимые для запросов реквизиты будут получены из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Пути к конфигу и логу прописаны в settings.json.
     *
     * @param array $fields Поля задачи (ассоциативный массив 'поле' => 'значение').
     *                      Подробно поля описаны в документации Битрикс24:
     *                      https://dev.1c-bitrix.ru/rest_help/tasks/task/tasks/tasks_task_add.php
     *
     * @return array|bool Ассоциативный массив с декодированными данными ответа
     *                    (в поле result),
     *                    либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function createTask(array $fields) {
        $result = Utils::doRequestToCrmViaModule(
            'tasks.task.add',
            'POST',
            array('fields' => $fields));

        if (($result === false) || ($result['success'] === false)) {
            $err = 'execution failed';
            if (isset($result['result']['error'])) {
                $err = $result['result']['error'];
            }
            Utils::logError(__METHOD__, sprintf(
                "Request failed: %s. Fields: %s",
                $err,
                json_encode($fields, JSON_UNESCAPED_UNICODE)));
            return false;
        }

        Utils::logDebug(__METHOD__,
            'Created task ID: '.$result['result']['task']['id']);

        return $result;
    }

    /**
     * Создать лид в Битрикс
     *
     * Поля и параметры подробно описаны в документации REST API Битрикс24:
     * https://dev.1c-bitrix.ru/rest_help/crm/leads/crm_lead_add.php
     *
     * Все необходимые для запросов реквизиты будут получены из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Пути к конфигу и логу прописаны в settings.json.
     *
     * @param array $fields Поля лида (ассоциативный массив 'поле' => 'значение').
     * @param array $params Набор параметров (ассоциативный массив 'поле' => 'значение').
     *
     * @return array|bool Ассоциативный массив с декодированными данными ответа
     *                    (в поле result),
     *                    либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function createLead(array $fields, array $params) {
        $result = Utils::doRequestToCrmViaModule(
            'crm.lead.add',
            'POST',
            array(
                'fields' => $fields,
                'params' => $params,
            ));

        if (($result === false) || ($result['success'] === false)) {
            $err = 'execution failed';
            if (isset($result['result']['error'])) {
                $err = $result['result']['error'];
            }
            Utils::logError(__METHOD__, sprintf(
                "Request failed: %s. Fields: %s; params: %s",
                $err,
                json_encode($fields, JSON_UNESCAPED_UNICODE),
                json_encode($params, JSON_UNESCAPED_UNICODE)));
            return false;
        }

        Utils::logDebug(__METHOD__,
            'Created lead ID: '.$result['result']);

        return $result;
    }

    /**
     * Обновить лид в Битрикс
     *
     * Поля и параметры подробно описаны в документации REST API Битрикс24:
     * https://dev.1c-bitrix.ru/rest_help/crm/leads/crm_lead_update.php
     *
     * Все необходимые для запросов реквизиты будут получены из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Пути к конфигу и логу прописаны в settings.json.
     *
     * @param int   $id     Идентификатор лида в Битрикс
     * @param array $fields Поля лида (ассоциативный массив 'поле' => 'значение').
     * @param array $params Набор параметров (ассоциативный массив 'поле' => 'значение').
     *
     * @return array|bool Ассоциативный массив с декодированными данными ответа
     *                    (в поле result),
     *                    либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function updateLead($id, array $fields, array $params) {
        $result = Utils::doRequestToCrmViaModule(
            'crm.lead.update',
            'POST',
            array(
                'id'     => $id,
                'fields' => $fields,
                'params' => $params,
            ));

        if (($result === false) || ($result['success'] === false)) {
            $err = 'execution failed';
            if (isset($result['result']['error'])) {
                $err = $result['result']['error'];
            }
            Utils::logError(__METHOD__, sprintf(
                "Request failed: %s. ID: %s; fields: %s; params: %s",
                $err,
                $id,
                json_encode($fields, JSON_UNESCAPED_UNICODE),
                json_encode($params, JSON_UNESCAPED_UNICODE)));
            return false;
        }

        if ($result['result'] === true) {
            Utils::logDebug(__METHOD__,
                'Successfully updated lead ID: '.$id);
        } else {
            Utils::logError(__METHOD__,
                'Failed to updated lead ID: '.$id);
        }

        return $result;
    }

    /**
     * Обновить контакт в Битрикс
     *
     * Поля и параметры подробно описаны в документации REST API Битрикс24:
     * https://dev.1c-bitrix.ru/rest_help/crm/contacts/crm_contact_update.php
     *
     * Все необходимые для запросов реквизиты будут получены из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Пути к конфигу и логу прописаны в settings.json.
     *
     * @param int   $id     Идентификатор лида в Битрикс
     * @param array $fields Поля контакта (ассоциативный массив 'поле' => 'значение').
     * @param array $params Набор параметров (ассоциативный массив 'поле' => 'значение').
     *
     * @return array|bool Ассоциативный массив с декодированными данными ответа
     *                    (в поле result),
     *                    либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function updateContact($id, array $fields, array $params) {
        $result = Utils::doRequestToCrmViaModule(
            'crm.contact.update',
            'POST',
            array(
                'id'     => $id,
                'fields' => $fields,
                'params' => $params,
            ));

        if (($result === false) || ($result['success'] === false)) {
            $err = 'execution failed';
            if (isset($result['result']['error'])) {
                $err = $result['result']['error'];
            }
            Utils::logError(__METHOD__, sprintf(
                "Request failed: %s. ID: %s; fields: %s; params: %s",
                $err,
                $id,
                json_encode($fields, JSON_UNESCAPED_UNICODE),
                json_encode($params, JSON_UNESCAPED_UNICODE)));
            return false;
        }

        if ($result['result'] === true) {
            Utils::logDebug(__METHOD__,
                'Successfully updated contact ID: '.$id);
        } else {
            Utils::logError(__METHOD__,
                'Failed to updated contact ID: '.$id);
        }

        return $result;
    }

    /**
     * Создать сделку в Битрикс
     *
     * Поля и параметры подробно описаны в документации REST API Битрикс24:
     * https://dev.1c-bitrix.ru/rest_help/crm/cdeals/crm_deal_add.php
     *
     * Все необходимые для запросов реквизиты будут получены из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Пути к конфигу и логу прописаны в settings.json.
     *
     * @param array $fields Поля сделки (ассоциативный массив 'поле' => 'значение').
     * @param array $params Набор параметров (ассоциативный массив 'поле' => 'значение').
     *
     * @return array|bool Ассоциативный массив с декодированными данными ответа
     *                    (в поле result),
     *                    либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function createDeal(array $fields, array $params) {
        $result = Utils::doRequestToCrmViaModule(
            'crm.deal.add',
            'POST',
            array(
                'fields' => $fields,
                'params' => $params,
            ));

        if (($result === false) || ($result['success'] === false)) {
            $err = 'execution failed';
            if (isset($result['result']['error'])) {
                $err = $result['result']['error'];
            }
            Utils::logError(__METHOD__, sprintf(
                "Request failed: %s. Fields: %s; params: %s",
                $err,
                json_encode($fields, JSON_UNESCAPED_UNICODE),
                json_encode($params, JSON_UNESCAPED_UNICODE)));
            return false;
        }

        Utils::logDebug(__METHOD__,
            'Created deal ID: '.$result['result']);

        return $result;
    }

    /**
     * Запустить бизнес-процесс в Битрикс
     *
     * Поля и параметры подробно описаны в документации REST API Битрикс24:
     * https://dev.1c-bitrix.ru/rest_help/bizproc/bizproc_workflow_instance/bizproc_workflow_start.php
     *
     * Все необходимые для запросов реквизиты будут получены из конфига.
     * При возникновении ошибки будет запись в лог модуля Itgrix.
     * Пути к конфигу и логу прописаны в settings.json.
     *
     * @param int   $templateId Идентификатор шаблона БП
     * @param array $documentId Идентификатор документа БП
     *                          (ассоциативный массив 'поле' => 'значение').
     *                          Пример: ['crm', 'CCrmDocumentLead', leadId]
     * @param array $parameters Значения параметров БП (если шаблон с параметрами)
     *                          (ассоциативный массив 'поле' => 'значение').
     *
     * @return array|bool Ассоциативный массив с декодированными данными ответа
     *                    (в поле result),
     *                    либо false в случае ошибки.
     *
     * @throws Exception В случае ошибки
     */
    public static function startBizproc($templateId, array $documentId, array $parameters) {
        $result = Utils::doRequestToCrmViaModule(
            'bizproc.workflow.start',
            'POST',
            array(
                'TEMPLATE_ID' => $templateId,
                'DOCUMENT_ID' => $documentId,
                'PARAMETERS'  => $parameters,
            ));

        if (($result === false) || ($result['success'] === false)) {
            $err = 'execution failed';
            if (isset($result['result']['error'])) {
                $err = $result['result']['error'];
            }
            Utils::logError(__METHOD__, sprintf(
                'Request failed: %s. Template ID: %s; document ID: %s; parameters: %s',
                $err,
                $templateId,
                json_encode($documentId, JSON_UNESCAPED_UNICODE),
                json_encode($parameters, JSON_UNESCAPED_UNICODE)));
            return false;
        }

        Utils::logDebug(__METHOD__,
            'Started bizproc workflow ID: '.$result['result']);

        return $result;
    }

}
