#!/usr/bin/env bash
{
set -o nounset

SCRIPT_DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"
readonly SCRIPT_DIR
SCRIPT_FILENAME="$(basename "${BASH_SOURCE[0]}")"
readonly SCRIPT_FILENAME

. "${SCRIPT_DIR}/include.sh"

readonly SCRIPT_NAME="${PRODUCT_NAME}-install"
readonly SCRIPT_VERSION="3.11.10"

LAUNCH_DATETIME_HUMAN="$(date "+${DATETIME_FORMAT_HUMAN}")"
readonly LAUNCH_DATETIME_HUMAN

SCRIPT_USAGE="$(cat <<EOF
Usage:
  ${SCRIPT_FILENAME} [options]

Options:
  --user=USER
  --group=GROUP
      User and group for application.

      App directory, config directory, and log file will be owned by USER:GROUP.
      Service will also run as USER:GROUP.

      Arguments MUST be either both specified or both omitted.
      By default, the user who runs Asterisk is used (this is usually "asterisk:asterisk").
      If the arguments are omitted and Asterisk is not detected,
      user and group will be requested interactively.
  --web-login=LOGIN
  --web-password=PASSWORD
      Login and password for web (admin panel).
      Neither login nor password can be empty.

      Arguments MUST be either both specified or both omitted.
      If the arguments are omitted, login and password will be requested interactively.
  --install-supervisor=true/false
      Install or not the ${PRODUCT_NAME}_supervisor service.
      By default, value is 'true'.
      The supervisor service is useful for applying updates in the admin panel.
  --existing-installation-mode=MODE
      What to do if existing installation found. MODE:
        1 - backup and reinstall
        2 - reinstall WITHOUT backup
        3 - Cancel and exit
      By default, installer will ask.
      WARNING:
        Configuration file will be reset to defaults during reinstallation.
        To save existing configuration file, select the backup option
        or cancel and make a manual backup.
  -v
      Show script version.
  --legal
      Show legal information.
  -h, --help
      Show help.
EOF
)"
readonly SCRIPT_USAGE

check_installation_exists() {
  if [ -d "${APP_DIR}" ] || [ -d "${CONFIG_DIR}" ]; then
    return 0
  fi
  return 1
}

check_installation_is_running() {
  if bin_is_running "${BIN_PATH}" || bin_is_running "${SUPERV_BIN_PATH}"; then
    return 0
  fi
  return 1
}

remove_installation() {
  local remover_path="${SCRIPT_DIR}/remove.sh"

  if [ -e "${REMOVE_EXE_PATH}" ]; then
    echo "Found existing installation remover, use it: '${REMOVE_EXE_PATH}'"
    remover_path="${REMOVE_EXE_PATH}"
  else
    echo "Existing installation remover not found: '${REMOVE_EXE_PATH}'"
    echo "Use new remover: '${remover_path}'"
  fi

  echo "Launch remover..."
  if ! "${remover_path}"; then
    return 1
  fi
  return 0
}

main() {
  host_user=
  host_group=
  ug_options="false"
  web_login=
  web_password=
  web_options="false"
  install_supervisor="true"
  existing_installation_mode='ask'

  for i in "$@"; do
    case "${i}" in
      --user=*)
        host_user="${i#*=}"
        ug_options="true"
        shift
        ;;
      --group=*)
        host_group="${i#*=}"
        ug_options="true"
        shift
        ;;
      --web-login=*)
        web_login="${i#*=}"
        web_options="true"
        shift
        ;;
      --web-password=*)
        web_password="${i#*=}"
        web_options="true"
        shift
        ;;
      --install-supervisor=*)
        install_supervisor="${i#*=}"
        shift
        ;;
      --existing-installation-mode=*)
        existing_installation_mode="${i#*=}"
        shift
        ;;
      -v)
        print_version
        exit
        shift
        ;;
      --legal)
        print_version
        echo
        print_legal
        exit
        shift
        ;;
      -h|--help)
        print_help
        exit
        shift
        ;;
      *) # unknown option
        die_usage "Unknown option: ${i}"
        ;;
    esac
  done

  print_version
  echo "Start at: ${LAUNCH_DATETIME_HUMAN}"

  # Check options
  echo
  echo "Check options..."
  if [ "${ug_options}" == "true" ] && \
     { [ "${host_user}" == "" ] || [ "${host_group}" == "" ]; }; then
    die_usage "$(gen_related_opts_error "--user" "--group")"
  fi

  if [ "${web_options}" == "true" ] && \
     { [ "${web_login}" == "" ] || [ "${web_password}" == "" ]; }; then
    die_usage "$(gen_related_opts_error "--web-login" "--web-password")"
  fi

  if [ "${install_supervisor}" != 'true' ] && [ "${install_supervisor}" != 'false' ]; then
    die_usage "'install-supervisor' option must be 'true' or 'false'"
  fi

  if [ "${existing_installation_mode}" != 'ask' ] && \
     ! is_uint_in_range "${existing_installation_mode}" 1 3
  then
    die_usage "Incorrect 'existing-installation-mode' option value: '${existing_installation_mode}'"
  fi
  echo "OK"

  # Check if current effective user is 'root'

  echo
  echo "Check current effective user..."

  local cur_eff_user
  local cur_eff_group
  cur_eff_user="$(id -un)"
  cur_eff_group="$(id -gn)"

  echo "Current effective user: '${cur_eff_user}:${cur_eff_group}'"

  if [ "${cur_eff_user}" != 'root' ]; then
    die "Current effective user is not 'root' (launch ${SCRIPT_NAME} as root or via sudo)"
  fi
  echo "OK"

  echo
  echo "********************************************************************************"
  echo "*** User and group for application"

  echo
  if [ "${ug_options}" == "true" ]; then
    echo "Check \"user\" and \"group\" options..."
    err_msg="$(check_user_group "${host_user}" "${host_group}")"
    ok_or_die $? "OK" "${err_msg}"
  else
    set_ug() {
      if [ -d "${PBX_LOG_DIR}" ]; then
        user="$(stat -c %U "${PBX_LOG_DIR}")"
        group="$(stat -c %G "${PBX_LOG_DIR}")"
        echo "${user}:${group}"
        return 0
      fi

      if check_user_group "${PBX_DEFAULT_USER}" "${PBX_DEFAULT_GROUP}" > /dev/null; then
        if prompt_yesno "Set user and group to ${PBX_DEFAULT_USER}:${PBX_DEFAULT_GROUP}?" yes ; then
          echo "${PBX_DEFAULT_USER}:${PBX_DEFAULT_GROUP}"
          return 0
        fi
      fi

      if prompt_yesno "Set user:group to \"root:root\"?" yes ; then
        echo "root:root"
        return 0
      fi

      local user=
      local group=

      while true; do
        read -p "Enter valid \"user:group\" (with colon): " ug
        user="${ug%:*}"
        group="${ug#*:}"
        msg="$(check_user_group "${user}" "${group}")"
        if [ $? -eq 0 ]; then
          echo "OK" 1>&2
          echo "${user}:${group}"
          return 0
        fi
        echo "${msg}" 1>&2
      done
    }
    ug_str="$(set_ug)"
    host_user="${ug_str%:*}"
    host_group="${ug_str#*:}"
  fi

  echo
  echo "User: ${host_user}"
  echo "Group: ${host_group}"

  echo
  echo "********************************************************************************"
  echo "*** Backup directory"

  echo
  if [ -d "${BACKUPS_DIR}" ]; then
    echo "Directory for backups already exists, do not create: '${BACKUPS_DIR}'"
  else
    echo "Create directory for backups: '${BACKUPS_DIR}'..."
    mkdir -p "${BACKUPS_DIR}"
    ok_or_die $? "OK" "Failed to create directory for backups: '${BACKUPS_DIR}'"
  fi

  echo
  echo "Set ownership of backup directory to '${host_user}:${host_group}': '${BACKUPS_DIR}'..."
  chown ${host_user}:${host_group} "${BACKUPS_DIR}"
  ok_or_die $? "OK" "Failed to set ownership of backup directory: '${BACKUPS_DIR}'"

  echo
  echo "********************************************************************************"
  echo "*** Check existing installation"

  local installation_exists='false'
  local old_version=''
  local old_is_running='false'

  echo
  echo "Search for existing '${PRODUCT_NAME}' installation or remaining files..."
  if check_installation_exists; then
    installation_exists='true'
    old_version="$(get_bin_version "${BIN_PATH}")"

    if check_installation_is_running; then
      old_is_running='true'
    fi
  fi

  if [ "${installation_exists}" != 'true' ]; then
    echo "Installation not found, install"
    existing_installation_mode='0'
  else
    echo -n "Found existing installation: version '${old_version}', "
    if [ "${old_is_running}" == 'true' ]; then
      echo 'RUNNING'
    else
      echo 'not running'
    fi
    echo
    echo "Existing installation mode: ${existing_installation_mode}"

    if [ "${existing_installation_mode}" == 'ask' ]; then
      local config_warning
      config_warning="$(cat << EOF
WARNING: Configuration file will be reset to defaults during reinstallation.
         To save current configuration file, select the backup option
         or cancel and make a manual backup.
EOF
      )"
      existing_installation_mode="$(prompt_numeric_select \
          "Select what to do:" \
          "${config_warning}" \
          3 1 \
          "Backup and reinstall" \
          "Reinstall WITHOUT backup" \
          "Cancel" \
          )"
    fi

    if [ "${existing_installation_mode}" == '3' ]; then # Cancel
      die "Cancelled by user"
    fi

    echo
    if [ "${existing_installation_mode}" == '1' ]; then # With backup
      "${SCRIPT_DIR}/backup.sh"
    else
      echo 'User selected option without backup, do not backup'
    fi

    echo
    echo "Remove existing installation..."
    if ! remove_installation; then
      die "Failed to remove existing installation"
    fi
    if check_installation_is_running; then
      die "Existing installation is still running after removing"
    fi

    echo
    echo "Previous installation removed, proceed new installation"
  fi

  echo
  echo "********************************************************************************"
  echo "*** Admin panel credentials"

  if [ "${web_options}" == "false" ]; then
    echo
    web_login="$(prompt_nonempty_value "Enter new login" \
                                       "Login can't be empty")"
    echo
    web_password="$(prompt_nonempty_value "Enter new password" \
                                          "Password can't be empty")"
  fi

  echo
  echo "Admin panel login: \"${web_login}\""
  echo "Admin panel password: \"${web_password}\""

  echo
  echo "********************************************************************************"
  echo "*** Prepare"

  echo
  echo "Create app directory..."
  mkdir -p "${APP_DIR}"
  ok_or_die $? "OK" "Failed to create dir: ${APP_DIR}"

  echo
  cd "${APP_DIR}" 2> /dev/null
  ok_or_die $? "Changed dir to ${APP_DIR}" "Failed to change dir: ${APP_DIR}"

  echo
  echo "Check permissions for ${APP_DIR} ..."
  check_dir_write_permissions "${APP_DIR}"
  ok_or_die $? "OK" "Permissions check failed: cannot write in ${APP_DIR}"

  # Локальный адрес модуля
  addr_array=($(ip addr show | grep -E -o "inet ([0-9]{1,3}[\.]){3}[0-9]{1,3}" | sed 's/inet //g'))
  internal_address="${addr_array[1]}"

  echo
  echo "********************************************************************************"
  echo "*** Install"

  echo
  echo "Remove old files, if any..."
  rm -rf "${BIN_FILENAME}"
  ok_or_die $? "OK" "Could not remove old files"

  echo
  echo "Copy new files..."
  cp -R "${SCRIPT_DIR}/"* "${APP_DIR}/"
  ok_or_die $? "OK" "Failed to copy new files"

  echo
  echo "Remove installer..."
  local installer_file_name=
  installer_file_name="$(basename "$0")"
  rm -f "${APP_DIR}/${installer_file_name}"
  ok_or_die $? "OK" "Failed to remove installer"

  echo
  echo "Initialize customizations..."
  cp -Rp ${CUSTOMIZER_DIR}{.dist,}
  ok_or_die $? "OK" "Failed to initialize customizations"

  echo
  echo "Create runtime files..."
  touch "${APP_DIR}/state.json" \
  && touch "${APP_DIR}/token.json"
  ok_or_die $? "OK" "Failed to create runtime files"

  echo
  echo "Set ownership..."
  chown --silent -R ${host_user}:${host_group} "${APP_DIR}"
  ok_or_die $? "OK" "Could not set ownership of new files"

  echo
  echo "********************************************************************************"
  echo "*** Configure"

  echo
  echo "Create config..."
  mkdir -p "${CONFIG_DIR}" \
  && cp "config.json.dist" "${CONFIG_PATH}" \
  && chown -R ${host_user}:${host_group} "${CONFIG_DIR}"
  ok_or_die $? "OK" "Failed to create config"

  echo
  echo "Load jedit..."
  . "${APP_DIR}/jedit/jedit_lib.sh"
  ok_or_die $? "OK" "Failed to load jedit"

  echo
  echo "Edit config..."
  jSet "${CONFIG_PATH}" "bitrix.recording_address" "\"${internal_address}\"" \
  && jSet "${CONFIG_PATH}" "server.login" "\"${web_login}\"" \
  && jSet "${CONFIG_PATH}" "server.password" "\"${web_password}\""
  ok_or_die $? "OK" "Failed to edit config"

  echo
  echo "********************************************************************************"
  echo "*** Log"

  echo
  echo "Create log: ${LOG_PATH} ..."
  touch "${LOG_PATH}" \
  && chown ${host_user}:${host_group} "${LOG_PATH}"
  ok_or_die $? "OK" "Failed to create log"

  logdir_u="$(stat -c %U "${LOG_DIR}")"
  logdir_g="$(stat -c %G "${LOG_DIR}")"

  if [ "${logdir_u}" == "" ] || [ "${logdir_g}" == "" ]; then
    die "Could not get owner of ${LOG_DIR}"
  fi

  echo
  echo "Create logrotate config..."
  cat <<EOF > "${LOGROTATE_CONFIG_PATH}"
${LOG_PATH} {
  su ${logdir_u} ${logdir_g}

  # ротация раз в день. Возможные варианты daily, weekly, monthly, size (например size=1M)
  daily

  # сохраняется последние 5 ротированных файла
  rotate 5

  # сжимать ротируемый файл
  compress

  # сжимать предыдущий файл при следующей ротации
  delaycompress

  # отсутствие файла не является ошибкой
  missingok

  # не создаём новый файл, копируем содержимое и обнуляем старый
  copytruncate
}
EOF
  ok_or_die $? "OK" "Failed to create logrotate config"

  echo
  check_logrotate "${LOGROTATE_CONFIG_PATH}" "${LOG_PATH}"

  echo
  echo "********************************************************************************"
  echo "*** Service"

  echo
  echo "Detect service manager..."

  if command -v systemctl &> /dev/null; then
    echo "Use systemctl" # модные

    # Проверяем, активны ли в системе необходимые сервисы
    db=
    if systemctl_is_active "mysql" ; then
      db="mysql"
    elif systemctl_is_active "mariadb" ; then
      db="mariadb"
    elif systemctl_is_active "mysqld" ; then
      db="mysqld"
    fi

    pbx=
    if systemctl_is_active "asterisk" ; then
      pbx="asterisk"
    elif systemctl_is_active "freepbx" ; then
      pbx="freepbx"
    fi

    systemd_service_file="/etc/systemd/system/${SERVICE_NAME}.service"

    echo
    echo "Create service config file..."
    touch "${systemd_service_file}"
    ok_or_die $? "OK" "Failed to create service config file"

    cat <<EOF > "${systemd_service_file}"
[Unit]
Description=${SERVICE_NAME}
EOF
    if [ "${db}" != "" ]; then
      cat <<EOF >> "${systemd_service_file}"
Requires=${db}.service
After=${db}.service
EOF
    fi
    if [ "${pbx}" != "" ]; then
      cat <<EOF >> "${systemd_service_file}"
Requires=${pbx}.service
After=${pbx}.service
EOF
    fi
    cat <<EOF >> "${systemd_service_file}"

[Service]
Type=simple
User=${host_user}
Group=${host_group}
ExecStart=/bin/sh -c "${BIN_PATH} >> ${LOG_PATH}"
Restart=on-failure

[Install]
WantedBy=multi-user.target
EOF

    echo
    echo "Enable service..."
    systemctl enable ${SERVICE_NAME}
    ok_or_die $? "OK" "Failed to enable service"

  elif command -v chkconfig &> /dev/null; then
    echo "Use chkconfig" # redhat

    service_script_path="/etc/init.d/${SERVICE_NAME}"

    echo
    echo "Copy service script..."
    cp "${APP_DIR}/service.sh" "${service_script_path}"
    ok_or_die $? "OK" "Failed to copy service script"

    # Modify service
    sed -E "s/\bUSERNAME=.*/USERNAME=${host_user}/" -i "${service_script_path}"

    echo
    echo "Add service..."
    chkconfig --add ${SERVICE_NAME}
    ok_or_die $? "OK" "Failed to add service"

  elif command -v update-rc.d &> /dev/null; then
    echo "Use update-rc.d" # debian

    service_script_path="/etc/init.d/${SERVICE_NAME}"

    echo
    echo "Copy service script..."
    cp "${APP_DIR}/service.sh" "${service_script_path}"
    ok_or_die $? "OK" "Failed to copy service script"

    # Modify service
    sed -E "s/\bUSERNAME=.*/USERNAME=${host_user}/" -i "${service_script_path}"

    echo
    echo "Add service..."
    update-rc.d ${SERVICE_NAME} defaults
    ok_or_die $? "OK" "Failed to add service"

  else
    die "Unknown service manager"
  fi

  echo
  echo "********************************************************************************"
  echo "*** Instance map"

  echo
  echo "Register instances..."
  "${BIN_PATH}" --register-instances
  ok_or_die $? "OK" "Failed to register instances"

  echo
  echo "********************************************************************************"
  echo "*** Sudoers"

  echo
  echo "Trying to use sudoers in order to enable update through admin panel"

  echo
  echo "Make and install sudoers file..."
  "${APP_DIR}/install_sudoers.sh"
  ok_or_die $? "OK" "Failed to install sudoers file"

  echo
  echo "********************************************************************************"
  echo "*** Supervisor"

  echo
  if [ "${install_supervisor}" == 'true' ]; then
    echo "Launch ${PRODUCT_NAME}_supervisor installer..."
    "${APP_DIR}/supervisor.dist/install.sh"
    ok_or_die $? "OK" "${PRODUCT_NAME}_supervisor installation failed"
  else
    echo "'install-supervisor' option is set to 'false', do not install ${PRODUCT_NAME}_supervisor"
  fi

  echo
  echo "********************************************************************************"
  echo "*** Finish"

  echo
  echo "Check the installed version..."
  get_bin_version "${BIN_PATH}"
  ok_or_die $? "OK" "Failed to check version"

  echo
  echo "Start service..."
  start_service "${SERVICE_NAME}"
  ok_or_die $? "OK" "Failed to start service"

  echo
  echo "Wait 1 sec and check service status for '${SERVICE_NAME}'..."
  sleep 1
  check_service_status "${SERVICE_NAME}"
  ok_or_die $? "OK" "Status check failed for service '${SERVICE_NAME}'"

  echo
  echo_box_h_border ' ╔' '═' '╗ '
  echo_boxed ' ║' '' '║ '
  echo_boxed ' ║' "Follow the link to complete the installation:" '║ '

  for i in "${!addr_array[@]}"; do
    if [ "${addr_array[$i]}" != "127.0.0.1" ]; then
      echo_boxed ' ║' '' '║ '
      echo_boxed ' ║' "$(print_link "http://${addr_array[$i]}:8077/config/master")" '║ '
    fi
  done

  echo_boxed ' ║' '' '║ '
  echo_boxed ' ║' "Login: ${web_login}" '║ '
  echo_boxed ' ║' "Password: ${web_password}" '║ '
  echo_boxed ' ║' '' '║ '
  echo_box_h_border ' ╚' '═' '╝ '

  echo -n "Finish " && print_version
}
main "$@"
exit
}
