#!/usr/bin/env bash
{
set -o nounset

SCRIPT_DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"
readonly SCRIPT_DIR
SCRIPT_FILENAME="$(basename "${BASH_SOURCE[0]}")"
readonly SCRIPT_FILENAME

. "${SCRIPT_DIR}/include.sh"

readonly SCRIPT_NAME="${PRODUCT_NAME}-install_sudoers"
readonly SCRIPT_VERSION="3.11.10"

LAUNCH_DATETIME_HUMAN="$(date "+${DATETIME_FORMAT_HUMAN}")"
readonly LAUNCH_DATETIME_HUMAN

SCRIPT_USAGE="$(cat <<EOF
Installs sudoers file for ${PRODUCT_NAME}.
LAUNCH ONLY AFTER INSTALLATION OF ${PRODUCT_NAME}!
RUN ONLY AS root (or via sudo)!

Usage:
  ${SCRIPT_FILENAME} [options]

Options:
  -v
      Show script version.
  --legal
      Show legal information.
  -h, --help
      Show help.
EOF
)" #"
readonly SCRIPT_USAGE


# WRITES TO GLOBALS:
#    commands
#
function addCommand() {
  local bin="$1"
  local arg_str="$2"

  if [ "${bin}" != '' ]; then
    commands+=("${bin} ${arg_str}")
  fi
}

function addCommands() {
  local coms=("$@")

  local i
  for ((i = 0; i < ${#coms[@]}; i += 2)); do
    addCommand "${coms[$i]}" "${coms[$((i+1))]}"
  done
}

function addMaintenanceCommands() {
  local cp_bin;        cp_bin="$(which cp)"
  local mkdir_bin;     mkdir_bin="$(which mkdir)"
  local chown_bin;     chown_bin="$(which chown)"
  local service_bin;   service_bin="$(which service)"
  local systemctl_bin; systemctl_bin="$(which systemctl)"

  addCommands \
    "${mkdir_bin}" "-p ${PRODUCT_TMP_DIR}" \
    "${mkdir_bin}" "-p ${BACKUPS_DIR}" \
    "${chown_bin}" "-R ${user}\:${group} ${PRODUCT_TMP_DIR}" \
    "${chown_bin}" "${user}\:${group} ${BACKUPS_DIR}" \
    "${chown_bin}" "-R ${user}\:${group} ${BACKUPS_DIR}" \
    "${chown_bin}" "-R ${user}\:${group} ${APP_DIR}" \
    "${cp_bin}"    "* ${BACKUPS_DIR}/*" \
  ;

  local service_name
  for service_name in "${service_names[@]}"; do
    addCommands \
      "${service_bin}" "${service_name} status" \
      "${service_bin}" "${service_name} stop" \
      "${service_bin}" "${service_name} start" \
      "${service_bin}" "${service_name} restart" \
    ;

    # will add only if systemctl_bin != ''
    addCommands \
      "${systemctl_bin}" "status ${service_name}" \
      "${systemctl_bin}" "stop ${service_name}" \
      "${systemctl_bin}" "start ${service_name}" \
      "${systemctl_bin}" "restart ${service_name}" \
    ;
  done

  addCommands "${service_bin}"   "${supervisor_service_name} status"

  # will add only if systemctl_bin != ''
  addCommands "${systemctl_bin}" "status ${supervisor_service_name}"
}

function addBugreportCommands() {
  local ls_bin;         ls_bin="$(which ls)"
  local tail_bin;       tail_bin="$(which tail)"
  local grep_bin;       grep_bin="$(which grep)"
  local journalctl_bin; journalctl_bin="$(which journalctl)"
  local chkconfig_bin;  chkconfig_bin="$(which chkconfig)"
  local more_bin;       more_bin="$(which more)"
  local crontab_bin;    crontab_bin="$(which crontab)"

  local mysql_bin;      mysql_bin="$(which mysql)"
  # mysql can be installed later
  mysql_bin="${mysql_bin:-/usr/bin/mysql}"

  addCommands \
    "${grep_bin}"       '* /var/log/messages' \
    "${grep_bin}"       '* /var/log/syslog' \
    "${ls_bin}"         "-la /etc/rc*/*${supervisor_service_name}" \
    "${journalctl_bin}" "-u ${supervisor_service_name}" \
    "${chkconfig_bin}"  "--list ${supervisor_service_name}" \
    "${tail_bin}"       '* /var/log/asterisk/full' \
    "${mysql_bin}"      "-e SHOW\\ VARIABLES\\ LIKE\\ \"%version%\";" \
    "${ls_bin}"         "-lia /var/spool/cron/crontabs/" \
		"${more_bin}"       "/var/spool/cron/crontabs/*" \
		"${crontab_bin}"    "-l" \
		"${crontab_bin}"    "-l -u asterisk" \
  ;

  local service_name
  for service_name in "${service_names[@]}"; do
    addCommands \
      "${ls_bin}"         "-la /etc/rc*/*${service_name}" \
      "${chkconfig_bin}"  "--list ${service_name}" \
      "${journalctl_bin}" "-u ${service_name}" \
    ;
  done
}

function printSudoersCmdAlias() {
  local alias_name="$1"
  local commands=("${@:2}")

  echo "Cmnd_Alias ${alias_name} = \\"

  local i
  for ((i = 0; i < ${#commands[@]}; ++i)); do
    echo -n "  ${commands[$i]}"

    if ((i != ${#commands[@]} - 1)); then
      echo ", \\"
    else
      echo
    fi
  done
}

function makeSudoersDraft() {
  local user="$1"
  local group="$2"
  local service_names=("${@:3}")

  local supervisor_service_name="${PRODUCT_NAME}_supervisor"
  local commands

  addCommand "$(which echo)" "check_${PRODUCT_NAME}_sudo"
  addCommand "$(which cat)" "/etc/sudoers.d/${PRODUCT_NAME}"

  addMaintenanceCommands
  addBugreportCommands

  local alias_name
  alias_name="$(echo "${PRODUCT_NAME}" | tr "[:lower:]" "[:upper:]")_CMD"
  printSudoersCmdAlias "${alias_name}" "${commands[@]}"
  printf "%s\tALL=(ALL:ALL)\tNOPASSWD: %s\n" "${user}" "${alias_name}"
}

function stagePrepare() {
  echo
  echo "Include jq from '${APP_DIR}/jq'..."
  . "${APP_DIR}/jq/jq_export.sh"
  ok_or_die $? "OK" "Failed to include jq"
  echo "jq version: $(jq --version)"

  echo
  echo "Check installation of ${PRODUCT_NAME} ..."
  if [ ! -e "${BIN_PATH}" ];  then
    die "'${PRODUCT_NAME}' is not (properly) installed: check executable at '${BIN_PATH}'"
  elif ! (jq '.[]' > /dev/null 2>&1 < "${INSTANCEMAP_PATH}"); then
    die "'${PRODUCT_NAME}' is not (properly) installed: check instance map at '${INSTANCEMAP_PATH}'"
  fi
  echo "OK"

  echo
  echo "Get owner:group of ${PRODUCT_NAME}..."
  host_user="$(stat -c %U "${BIN_PATH}")"
  host_group="$(stat -c %G "${BIN_PATH}")"
  echo "OK"
  echo "User: ${host_user}"
  echo "Group: ${host_group}"
}

function stageCheck() {
  echo
  echo "Check selected user..."
  if [ "${host_user}" == "root" ]; then
    echo "Selected user is 'root', do not use sudoers"
    return 1

  elif ! which sudo > /dev/null || ! which visudo > /dev/null; then

    warn "$(cat << EOF
You will not be able to update through admin panel,
  because selected user to run ${PRODUCT_NAME} is not 'root' ('${host_user}')
  and 'sudo' command was not found.
  Install sudo and reinstall ${PRODUCT_NAME}, or contact our technical support.
EOF
)"
    print_tech_sup_info
    return 1

  elif ! [ -d "${SUDOERS_DIR}" ]; then

    warn "$(cat << EOF
You will not be able to update through admin panel,
  because selected user to run ${PRODUCT_NAME} is not 'root' ('${host_user}')
  and sudoers directory was not found: '${SUDOERS_DIR}'
EOF
)"
    print_tech_sup_info
    return 1
  fi
  return 0
}

function stageInstall() {
  echo "Make sudoers file for selected user: '${host_user}'"

  echo
  echo "Get services from instancemap..."
  local services
  # shellcheck disable=SC2207
  services=($(jq -r '.[] | .service' < "${INSTANCEMAP_PATH}"))
  ok_or_die $? "OK" "Failed to parse instancemap: '${INSTANCEMAP_PATH}'"

  echo
  echo "Registered services in instancemap:"
  (IFS=,; echo "${services[*]}" | tr , '\n')

  echo
  echo "Make sudoers draft file: '${SUDOERS_DRAFT_FILE}'..."
  makeSudoersDraft "${host_user}" "${host_group}" "${services[@]}" > "${SUDOERS_DRAFT_FILE}"
  ok_or_die $? "OK" "Failed to make sudoers draft file: '${SUDOERS_DRAFT_FILE}'"

  echo
  echo "Set sudoers draft file permissions..."
  chmod 0440 "${SUDOERS_DRAFT_FILE}"
  ok_or_die $? "OK" "Failed to set sudoers draft file permissions: '${SUDOERS_DRAFT_FILE}'"

  echo
  echo "Check sudoers draft file: '${SUDOERS_DRAFT_FILE}'..."
  visudo -cs -f "${SUDOERS_DRAFT_FILE}"
  ok_or_die $? "OK" "Sudoers draft file check failed: '${SUDOERS_DRAFT_FILE}'"

  echo
  echo "Apply sudoers draft..."
  cp "${SUDOERS_DRAFT_FILE}" "${SUDOERS_FILE}"
  ok_or_die $? "OK" "Sudoers draft check failed: '${SUDOERS_DRAFT_FILE}'"

  echo
  echo "Check sudoers..."
  echo "check_${PRODUCT_NAME}_sudo" > /dev/null
  ok_or_die $? "OK" "Sudoers check failed"

  echo
  echo "Remove sudoers draft file: '${SUDOERS_DRAFT_FILE}'..."
  rm -f "${SUDOERS_DRAFT_FILE}"
  ok_or_die $? "OK" "Failed to remove sudoers draft file: '${SUDOERS_DRAFT_FILE}'"
}

function parseArgs() {
  local arg
  for arg in "$@"; do
    case "${arg}" in
      -v)
        print_version
        exit 0
        ;;
      --legal)
        print_version
        echo
        print_legal
        exit 0
        ;;
      -h|--help)
        print_help
        exit 0
        ;;
      *)
        die_usage "Unknown option: ${arg}"
        ;;
    esac
  done
}

function main() {
  parseArgs "$@"

  print_version
  echo "Start at: ${LAUNCH_DATETIME_HUMAN}"

  local host_user
  local host_group

  stagePrepare

  if stageCheck; then
    stageInstall
  fi

  echo
  echo -n "Finish " && print_version
}
main "$@"
exit
}
