#!/usr/bin/env bash

CUR_DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"

. "${CUR_DIR}/include.sh"

. "${APP_DIR}/jq/jq_export.sh"
. "${APP_DIR}/jedit/jedit_lib.sh"

foreach_instance() {
  local func="$1"

  echo "Start processing instances (with '${func}')..."

  local instances
  instances=($(cat "${INSTANCEMAP_PATH}" | jq -c '.[]'))

  echo "Got ${#instances[@]} instances from instancemap"

  local instance
  local index
  local suffix
  local config
  local service
  local log
  for instance in "${instances[@]}"; do

    index="$(echo "${instance}" | jq -r '.index')"
    config="$(echo "${instance}" | jq -r '.config')"
    service="$(echo "${instance}" | jq -r '.service')"
    log="$(echo "${instance}" | jq -r '.log')"

    suffix="-${index}"
    if [ "${suffix}" == '-0' ]; then
      suffix=''
    fi

    echo "Processing instance #${index}..."
    echo "  (config: '${config}')"
    echo "  (service: '${service}')"
    echo "  (log: '${log}')"

    "${func}" \
        "${suffix}" \
        "${config}" \
        "${service}" \
        "${log}"

    echo "Processing instance #${index} completed"
  done

  echo "Processing of all instances completed"
}

# needs errexit option
reinstall_sudoers() {
  echo "Reinstall sudoers file, if installed"
  if ! [ -f "${SUDOERS_FILE}" ]; then
    echo "Sudoers file is not installed, do nothing"
    return 0
  fi

  echo "Remove old sudoers file..."
  "${APP_DIR}/remove_sudoers.sh"
  echo 'OK'

  echo "Install new sudoers file..."
  "${APP_DIR}/install_sudoers.sh"
  echo 'OK'

  echo "Done: sudoers file reinstalled."
}

update_customizer_libs() {
  local customizerDir="$1"

  echo "Update customizer libs in '${customizerDir}'..."
  if ! [ -d "${customizerDir}" ]; then
    echo "ERROR: Customizer directory does not exist: '${customizerDir}'"
    return 1
  fi
  if ! [ -d "${APP_DIR}/customizer.dist" ]; then
    echo "ERROR: No 'customizer.dist' directory in '${APP_DIR}'"
    return 1
  fi

  local files=(\
    "customize.php" \
    "lib/crm.php" \
    "lib/Customizer.php" \
    "lib/Settings.php" \
    "lib/Utils.php" \
  )

  for file in "${files[@]}"; do
    cp -vp "${CUR_DIR}/customizer.dist/${file}" "${customizerDir}/${file}"
  done

  echo "Customizer libs updated in '${customizerDir}'"
}

# update_if_default accepts customizer path, action name and hash of expected old file contents.
# If action file hash matches expected old hash it gets replaced with new default file from dist.
update_if_default() {
  local customizer_dir="$1"
  local action="$2"
  local old_sum="$3"

  local action_file="${customizer_dir}/actions/${action}.php"

  if [ "$(md5sum < "${action_file}" || :)" == "${old_sum}" ]; then
    echo "'${action_file}' is unchanged, replace with new default..."
    cp -v "${APP_DIR}/customizer.dist/actions/${action}.php" "${action_file}"
  else
    echo "'${action_file}' has changes, keep current version"
  fi
}

# update_default_actions accepts customizer path,
# replaces default action files with new default files from dist if they match expected old hashes.
update_default_actions() {
  local customizer_dir="$1"

  echo "Update unchanged default actions in '${customizer_dir}'..."

  # These hashes must be updated manually when default action files are changed.
  update_if_default "${customizer_dir}" "customize_activity"           "51f9382ec18e1ec8b306cf34bf1a50e6  -"
  update_if_default "${customizer_dir}" "customize_autodistribution"   "9dfa2d0f7103ac541c53969fc2bb11c5  -"
  update_if_default "${customizer_dir}" "customize_entity"             "47ec7d8abc570f425ccee98b0080a0c5  -"
  update_if_default "${customizer_dir}" "need_registration"            "75982ce823d3f967e6e019ab3e1603a7  -"
  update_if_default "${customizer_dir}" "process_client_phone"         "3bde92e1de0176e3c48844455f24e45c  -"
  update_if_default "${customizer_dir}" "process_final_params"         "419917e914bd38335f606e976c56f7d6  -"
  update_if_default "${customizer_dir}" "process_originate_params"     "2edb931ea4e4ec419126ad2e7635a864  -"
  update_if_default "${customizer_dir}" "process_record_file_path"     "23277dd2efc99597f03e5e19444dbc2d  -"
  update_if_default "${customizer_dir}" "process_registration_params"  "422f09f15a804135def47f27daf2a624  -"
  update_if_default "${customizer_dir}" "process_task_creation_params" "4c3e6400a3c5f4ebb421f3212031b54f  -"
  update_if_default "${customizer_dir}" "process_user_phone"           "fd304b09c6eaf882a34efee9e0fa9f41  -"

  echo "Actions updated in '${customizer_dir}'"
}
