#!/usr/bin/env bash
{
set -o nounset

SCRIPT_DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"
readonly SCRIPT_DIR
SCRIPT_FILENAME="$(basename "${BASH_SOURCE[0]}")"
readonly SCRIPT_FILENAME

. "${SCRIPT_DIR}/include.sh"

readonly SCRIPT_NAME="${PRODUCT_NAME}-install"
readonly SCRIPT_VERSION="1.3.4"

LAUNCH_DATETIME_HUMAN="$(date "+${DATETIME_FORMAT_HUMAN}")"
readonly LAUNCH_DATETIME_HUMAN

SCRIPT_USAGE="$(cat <<EOF
LAUNCH ONLY AFTER INSTALLATION OF ${MAIN_PRODUCT_NAME}!

Usage:
  ${SCRIPT_FILENAME} [options]

Options:
  -v
      Show script version.
  --legal
      Show legal information.
  -h, --help
      Show help.
EOF
)"
readonly SCRIPT_USAGE

main() {

  for i in "$@"; do
    case "${i}" in
      -v)
        print_version
        exit
        shift
        ;;
      --legal)
        print_version
        echo
        print_legal
        exit
        shift
        ;;
      -h|--help)
        print_help
        exit
        shift
        ;;
      *) # unknown option
        die_usage "Unknown option: ${i}"
        ;;
    esac
  done

  print_version
  echo "Start at: ${LAUNCH_DATETIME_HUMAN}"

  echo
  echo "================================================================================"
  echo "=== Check installation of ${MAIN_PRODUCT_NAME}"

  echo
  echo "Get owner:group of ${MAIN_PRODUCT_NAME}..."
  if ! [ -e "${MAIN_BIN_PATH}" ]; then
    die "${MAIN_PRODUCT_NAME} is not (properly) installed in ${MAIN_APP_DIR}"
  fi

  local host_user="$(stat -c %U "${MAIN_BIN_PATH}")"
  local host_group="$(stat -c %G "${MAIN_BIN_PATH}")"

  echo
  echo "User: ${host_user}"
  echo "Group: ${host_group}"

  echo
  echo "Check ${MAIN_PRODUCT_NAME} version..."
  "${MAIN_BIN_PATH}" -v
  ok_or_die $? "OK" "Could not get ${MAIN_PRODUCT_NAME} version"

  echo
  echo "Include jq from '${MAIN_APP_DIR}/jq'..."
  . "${MAIN_APP_DIR}/jq/jq_export.sh"
  ok_or_die $? "OK" "Failed to include jq"
  echo "jq version: $(jq --version)"

  echo
  echo "================================================================================"
  echo "=== Install"

  echo
  echo "App directory: '${APP_DIR}'"

  echo
  echo "Remove old files, if any..."
  rm -rf "${APP_DIR}"
  ok_or_die $? "OK" "Could not remove old files"

  echo
  echo "Create app directory..."
  mkdir -p "${APP_DIR}"
  ok_or_die $? "OK" "Failed to create dir: ${APP_DIR}"

  echo
  echo "Copy new files..."
  cp -R "${SCRIPT_DIR}/"* "${APP_DIR}/"
  ok_or_die $? "OK" "Failed to copy new files"

  echo
  echo "Set ownership..."
  chown --silent -R ${host_user}:${host_group} "${APP_DIR}"
  ok_or_die $? "OK" "Could not set ownership of new files"

  echo
  echo "================================================================================"
  echo "=== Config"

  echo
  echo "Check config..."
  if ! [ -f "${CONFIG_PATH}" ]; then
    warn "Config file does not exist: '${CONFIG_PATH}'"
  else
    echo "OK"
  fi

  echo
  echo "================================================================================"
  echo "=== Log"

  echo
  echo "Create log: ${LOG_PATH} ..."
  touch "${LOG_PATH}" \
  && chown ${host_user}:${host_group} "${LOG_PATH}"
  ok_or_die $? "OK" "Failed to create log"

  logdir_u="$(stat -c %U "${LOG_DIR}")"
  logdir_g="$(stat -c %G "${LOG_DIR}")"

  if [ "${logdir_u}" == "" ] || [ "${logdir_g}" == "" ]; then
    die "Could not get owner of ${LOG_DIR}"
  fi

  echo
  echo "Create logrotate config..."
  cat <<EOF > "${LOGROTATE_CONFIG_PATH}"
${LOG_PATH} {
  su ${logdir_u} ${logdir_g}

  # ротация раз в день. Возможные варианты daily, weekly, monthly, size (например size=1M)
  daily

  # сохраняется последние 5 ротированных файла
  rotate 5

  # сжимать ротируемый файл
  compress

  # сжимать предыдущий файл при следующей ротации
  delaycompress

  # отсутствие файла не является ошибкой
  missingok

  # не создаём новый файл, копируем содержимое и обнуляем старый
  copytruncate
}
EOF
  ok_or_die $? "OK" "Failed to create logrotate config"

  echo
  check_logrotate "${LOGROTATE_CONFIG_PATH}" "${LOG_PATH}"

  echo
  echo "================================================================================"
  echo "=== Service"

  echo
  echo "Detect service manager..."

  if command -v systemctl &> /dev/null; then
    echo "Use systemctl" # модные

    systemd_service_file="/etc/systemd/system/${SERVICE_NAME}.service"

    echo
    echo "Create service config file..."
    touch "${systemd_service_file}"
    ok_or_die $? "OK" "Failed to create service config file"

    cat <<EOF > "${systemd_service_file}"
[Unit]
Description=${SERVICE_NAME}
[Service]
Type=simple
User=${host_user}
Group=${host_group}
ExecStart=/bin/sh -c "${BIN_PATH} >> ${LOG_PATH}"
Restart=on-failure

[Install]
WantedBy=multi-user.target
EOF

    echo
    echo "Enable service..."
    systemctl enable ${SERVICE_NAME}
    ok_or_die $? "OK" "Failed to enable service"

  elif command -v chkconfig &> /dev/null; then
    echo "Use chkconfig" # redhat

    service_script_path="/etc/init.d/${SERVICE_NAME}"

    echo
    echo "Copy service script..."
    cp "${APP_DIR}/service.sh" "${service_script_path}"
    ok_or_die $? "OK" "Failed to copy service script"

    # Modify service
    sed -E "s/\bUSERNAME=.*/USERNAME=${host_user}/" -i "${service_script_path}"

    echo
    echo "Add service..."
    chkconfig --add ${SERVICE_NAME}
    ok_or_die $? "OK" "Failed to add service"

  elif command -v update-rc.d &> /dev/null; then
    echo "Use update-rc.d" # debian

    service_script_path="/etc/init.d/${SERVICE_NAME}"

    echo
    echo "Copy service script..."
    cp "${APP_DIR}/service.sh" "${service_script_path}"
    ok_or_die $? "OK" "Failed to copy service script"

    # Modify service
    sed -E "s/\bUSERNAME=.*/USERNAME=${host_user}/" -i "${service_script_path}"

    echo
    echo "Add service..."
    update-rc.d ${SERVICE_NAME} defaults
    ok_or_die $? "OK" "Failed to add service"

  else
    die "Unknown service manager"
  fi

  echo
  echo "================================================================================"
  echo "=== Finish"

  echo
  echo "Check the installed ${PRODUCT_NAME} version..."
  "${BIN_PATH}" -v
  ok_or_die $? "OK" "Failed to check version"

  echo
  echo "Start service..."
  start_service "${SERVICE_NAME}"
  ok_or_die $? "OK" "Failed to start service"

  echo
  echo "Wait 1 sec and check service status for '${SERVICE_NAME}'..."
  sleep 1
  check_service_status "${SERVICE_NAME}"
  ok_or_die $? "OK" "Status check failed for service '${SERVICE_NAME}'"

  echo
  echo -n "Finish " && print_version
}
main "$@"
exit
}
